# TA implementation of nematic --> need pythonize this


# package Geometry;
#
# import Commons.Point3D;
# import Commons.Saver;
# import MyShapes.MyLine2D;
# import java.awt.Graphics2D;
# import java.awt.Point;
# import java.awt.geom.Point2D;
# import java.awt.image.BufferedImage;
# import java.util.ArrayList;
#
# /**
#  * Nematic class
#  *
#  * @author Benoit Aigouy
#  */
import math
import sys
import traceback

from numba import jit, njit
from PIL.ImageDraw import ImageDraw
from skimage.draw import line_aa, line
import numpy as np
from skimage.measure import regionprops
from skimage.measure import label
from skimage.measure._regionprops import RegionProperties

from epyseg.draw.shapes.line2d import Line2D

#
#   /**
#      *
#      * @param stretch_rescaling_factor
#      * @return the components of the strecth nematic of the flooded area (and
#      * the coordinates of the nematic extremities in case it has to be plotted)
#      */
#     public Point2D.Double[] compute_stretch(double stretch_rescaling_factor) {
#         double count = 0.;
#
#         double S1 = 0.;
#         double S2 = 0.;
#
#         Point2D.Double center = getCenter2D();
# //COMPUTE NEMATIC TENSOR FOR THE STRETCH
#
#         for (int k = 0; k < queueSize; k++) {
#             //GET THE CURRENT POINT (CONSTANT INTENSITY FOR THE STRETCH)
#             //COMPUTE NECESSARY TRIGONOMETRIC QUANTITES FOR EACH POINT
#             double deltaX = (double) flood_x[k] - center.x; //ARE THESE CASTS NECESSARY?
#             double deltaY = (double) flood_y[k] - center.y;
#             double rSquared = deltaX * deltaX + deltaY * deltaY;
#             double cos2Theta = 2. * deltaX * deltaX / rSquared - 1.;
#             double sin2Theta = 2. * deltaX * deltaY / rSquared;
#
#             //ADD THE SUMMANDS TO THEIR RESPECTIVE SUMS...
#             S1 += cos2Theta;
#             S2 += sin2Theta;
#             count++;
#         }
#
#         S1 /= count;//we rescale by the area
#         S2 /= count;
#
#         //COMPUTE THE STRETCH NEMATIC
#         double S0 = Math.sqrt(S1 * S1 + S2 * S2);
#         double cos2ThetaS = S1 / S0;
#         double sin2ThetaS = S2 / S0;
#         double sinThetaS = Math.sqrt(0.5 * (1. - cos2ThetaS));
#         double cosThetaS = 0.5 * sin2ThetaS / sinThetaS;
#
# //IN THE END WE WANT TO DRAW A LINE THROUGH THE CENTER OF THE CELL REPRESENTING THE STRETCH NEMATIC
#         double scale = stretch_rescaling_factor * S0;
#
#         double xDem = center.x - scale * cosThetaS;
#         double xFin = center.x + scale * cosThetaS;
#         double yDem = center.y - scale * sinThetaS;
#         double yFin = center.y + scale * sinThetaS;
#
#         Point2D.Double[] out = new Point2D.Double[3];
#         out[0] = new Point2D.Double(xDem, yDem);
#         out[1] = new Point2D.Double(xFin, yFin);
#         out[2] = new Point2D.Double(S1, S2);
#         return out;
#     }

# Computes a stretch nematic based on cell area
from epyseg.img import Img


# def compute_stretch_nematic_regionprops_extra(regionmask, intensity_image=None):
#     print(type(regionmask))
#     print(regionmask.shape)# --> 11,39 --> not at all what I want
#     print(regionmask)
#
#
#     return compute_stretch_nematic(regionmask)


def compute_stretch_nematic(pointsWithinArea_or_region, normalizeByArea=True):
    if pointsWithinArea_or_region is None:
        return
    if isinstance(pointsWithinArea_or_region, RegionProperties):
        pointsWithinArea_or_region = pointsWithinArea_or_region.coords
    # center_x = 0
    # center_y = 0
    # counter = 0
    # for point in pointsWithinArea_or_region:
    #     center_x += point[1]
    #     center_y += point[0]
    #     counter += 1
    # center_x/=counter
    # center_y/=counter

    # print(center_x, center_y)


    # print(np.average(pointsWithinArea_or_region, axis=0))
    S1,S2,center = _compute_stretch_nematic(pointsWithinArea_or_region,normalizeByArea=normalizeByArea)
    return Nematic(S1=S1, S2=S2, center=center)

# @njit
def _compute_stretch_nematic(pointsWithinArea_or_region, normalizeByArea=True):
    # center_y = 0.
    # center_x=0.
    center_y, center_x = np.average(pointsWithinArea_or_region, axis=0)

    # center_x /= pointsWithinArea_or_region.shape[0]
    # center_y /= pointsWithinArea_or_region.shape[0]

    # print(center_x, center_y)

    center = [center_y, center_x]

    S1 = 0.
    S2 = 0.

    # I guess it's really time to recode this properly using numpy --> do that when I have time
    # MEGA TODO --> CONVERT IT TO NUMPY BUT OK FOR NOW
    # //COMPUTE NEMATIC TENSOR FOR THE STRETCH
    for point in pointsWithinArea_or_region:
        # //GET THE CURRENT POINT (CONSTANT INTENSITY FOR THE STRETCH)
        # //Point pt = points_unsrt.get(k);
        #
        # //COMPUTE NECESSARY TRIGONOMETRIC QUANTITES FOR EACH POINT

        deltaX = float(point[1]) - center_x  # //ARE THESE CASTS NECESSARY?
        deltaY = float(point[0]) - center_y
        rSquared = deltaX * deltaX + deltaY * deltaY
        cos2Theta = 2. * deltaX * deltaX / rSquared - 1.
        sin2Theta = 2. * deltaX * deltaY / rSquared

        # //ADD THE SUMMANDS TO THEIR RESPECTIVE SUMS...
        S1 += cos2Theta
        S2 += sin2Theta

    # //--> MAYBE REMOVE THIS TO MAKE IT SIZE DEPENDENT OR ADD A BOOLEAN
    if normalizeByArea:
        S1 /= pointsWithinArea_or_region.shape[0]
        S2 /= pointsWithinArea_or_region.shape[0]
    return S1,S2, center

# builds an average nematic from a series of nematics
# build from other nemats and/or their S1S2 components
# never tried --> really need do it!!!
# shall I offer a center ???
def compute_average_nematic(*args):
    if args is None or len(args) == 0:
        return
    S1 = 0
    S2 = 0
    if isinstance(args[0], Nematic):
        for nemat in args:
            S1 += nemat.S1
            S2 += nemat.S2
        S1 /= len(args)
        S2 /= len(args)
        return Nematic(S1=S1, S2=S2)
    else:
        for S1S2 in args:
            S1 += S1S2[0]
            S2 += S1S2[1]
        S1 /= len(args)
        S2 /= len(args)



class Nematic():

    # NB angle should be in radians not in degrees could maybe offer the option to have degrees too...
    def __init__(self, S1S2=None, S1=None, S2=None, S0=None, angle=None, base=None, tip=None,
                 center=None):  # TODO maybe some day use base and tip --> revive old TA CODE FOR IT or delete it
        self.rescaling_factor = 1.

        # create a null nematic
        self.S1 = 0
        self.S2 = 0
        self.center = [0, 0]

        if base is not None and tip is not None:
            # print('init',base, tip)
            from epyseg.ta.measurements.TAmeasures import distance_between_points
            center = ((base[0] + tip[0]) / 2., (tip[1] + base[1]) / 2.)
            angle = math.atan2(tip[0] - base[0], tip[1] - base[1])
            S0 = distance_between_points(center, tip)

        # if S1 and S2 are defined then use those
        if S1S2 is not None:
            self.S1 = S1S2[0]
            self.S2 = S1S2[1]
            self.center = [0, 0]
        if S1 is not None:
            self.S1 = S1
        if S2 is not None:
            self.S2 = S2
        if center is not None:
            self.center = center
        if angle is not None and S0 is not None:
            while angle < 0:
                angle += math.pi
            while angle >= math.pi:
                angle -= math.pi
            cos2Theta0 = math.cos(2. * angle);
            sin2Theta0 = math.sin(2. * angle);
            self.S1 = cos2Theta0 * S0
            self.S2 = sin2Theta0 * S0

    def set_rescaling_factor(self, rescaling_factor):
        self.rescaling_factor = rescaling_factor

    #     /**
    #      * This constructor converts a vector to a nematic
    #      *
    #      * @param v vector we want to convert to a nematic
    #      */
    #     public Nematic(Vector v) {
    #         this(v.getBase(), v.getTip());
    #     }
    #     /**
    #      * Nematic constructor
    #      *
    #      * @param center center of the nematic
    #      * @param any_extremity_pt coordinates of any extremity of the neamtic
    #      */
    #     public Nematic(Point2D.Double center, Point2D.Double any_extremity_pt) {
    #         this(center, any_extremity_pt.distance(center), Math.atan2(any_extremity_pt.y - center.y, any_extremity_pt.x - center.x));
    #     }
    #
    #     /**
    #      * Nematic constructor
    #      *
    #      * @param base
    #      * @param tip
    #      * @param inutile
    #      */
    #     public Nematic(Point2D.Double base, Point2D.Double tip, boolean inutile) {
    #         this(new Point2D.Double((base.x + tip.x) / 2., (tip.y + base.y) / 2.), tip);
    #     }
    #

    #
    #     public Nematic(ArrayList<Point2D.Double> S1S2) {
    #         this(S1S2.toArray(new Point2D.Double[S1S2.size()]));
    #     }
    #
    #     /**
    #      * builds an average nematic from a series of nematics
    #      *
    #      * @param S1andS2Components
    #      * @since <B>Tissue Analyzer 1.0</B>
    #      */
    #     public Nematic(Nematic... S1andS2Components) {
    #         if (S1andS2Components == null || S1andS2Components.length == 0) {
    #             return;
    #         }
    #         for (Nematic double1 : S1andS2Components) {
    #             S1 += double1.S1;
    #             S2 += double1.S2;
    #         }
    #         S1 /= S1andS2Components.length;
    #         S2 /= S1andS2Components.length;
    #     }
    #
    #     /**
    #      * builds an average nematic from a series of nematics
    #      *
    #      * @param S1andS2Components
    #      * @since <B>Tissue Analyzer 1.0</B>
    #      */
    #     public Nematic(Point2D.Double... S1andS2Components) {
    #         if (S1andS2Components == null || S1andS2Components.length == 0) {
    #             return;
    #         }
    #         for (Point2D.Double double1 : S1andS2Components) {
    #             S1 += double1.x;
    #             S2 += double1.y;
    #         }
    #         S1 /= (double) S1andS2Components.length;
    #         S2 /= (double) S1andS2Components.length;
    #     }
    #

    def getS0(self):
        return self.getMagnitude()

    # return the magnitude of the nematic
    def getMagnitude(self):
        return math.sqrt((self.S1) * (self.S1) + (self.S2) * (self.S2))

    # return the center of the nematic
    def getCenter(self):
        return self.center

    #
    def setCenter(self, center):
        self.center = center

    #
    #     /**
    #      *
    #      * @return a point3D containing S1, S2 and the magnitude of the nematic
    #      */
    def getS1S2S0(self):
        return [self.S1, self.S2, self.getS0()]

    # return the normal for the current nematic
    def getNormal(self):
        return Nematic(S1=-self.S1, S2=-self.S2, center=self.center)

    # return the components of the nematic
    def getS1S2(self):
        return [self.S1, self.S2]

    # TODO replace the complex acos original stuff below with that more elegant code
    def get_angle2(self):
        angle = 0.5 * math.atan2(self.S2, self.S1)
        while angle < 0:
            angle += math.pi
        while angle >= math.pi:
            angle -= math.pi
        return angle

    # return the orientation of the nematic in radians
    def getAngleInRadians(self):
        S0 = self.getMagnitude()
        cos2Theta0 = self.S1 / S0
        sinTheta0 = math.sqrt(0.5 * (1. - cos2Theta0))
        sin2Theta0 = self.S2 / S0
        cosTheta0 = 0.5 * sin2Theta0 / sinTheta0
        angle = math.acos(cosTheta0)
        while (angle < 0):
            angle += math.pi
        while (angle >= math.pi):
            angle -= math.pi
        return angle

    #
    #     /**
    #      *
    #      * @param rescaling_factor
    #      * @return the coordinates of the base of the nematic after rescaling
    #      */
    #     public Point2D.Double getBase(double rescaling_factor) {
    #         double S0 = Math.sqrt(S1 * S1 + S2 * S2);
    #         double cos2Theta0 = S1 / S0;
    #         double sin2Theta0 = S2 / S0;
    #         double sinTheta0 = Math.sqrt(0.5 * (1. - cos2Theta0));
    #         double cosTheta0 = 0.5 * sin2Theta0 / sinTheta0;
    #         double scale = rescaling_factor * S0;
    #         double xDem = center.x - scale * cosTheta0;
    #         double yDem = center.y - scale * sinTheta0;
    #         return new Point2D.Double(xDem, yDem);
    #     }
    #
    #     /**
    #      *
    #      * @param rescaling_factor
    #      * @return the coordinates of the tip of the nematic after rescaling
    #      */
    #     public Point2D.Double getTip(double rescaling_factor) {
    #         double S0 = Math.sqrt(S1 * S1 + S2 * S2);
    #         double cos2Theta0 = S1 / S0;
    #         double sin2Theta0 = S2 / S0;
    #         double sinTheta0 = Math.sqrt(0.5 * (1. - cos2Theta0));
    #         double cosTheta0 = 0.5 * sin2Theta0 / sinTheta0;
    #         double scale = rescaling_factor * S0;
    #         double xFin = center.x + scale * cosTheta0;
    #         double yFin = center.y + scale * sinTheta0;
    #         return new Point2D.Double(xFin, yFin);
    #     }
    #
    #     /**
    #      *
    #      * @return the coordinates of the base and tip of the nematic
    #      */
    #     public Point2D.Double[] getBeginAndEnd() {
    #         return getBeginAndEnd(1.);
    #     }
    #
    #     /**
    #      * @param rescaling_factor
    #      * @return the coordinates of the base and tip of the nematic after
    #      * rescaling
    #      */
    def getBeginAndEnd(self, rescaling_factor=1.):
        S0 = math.sqrt(self.S1 * self.S1 + self.S2 * self.S2)
        if S0 == 0:
            # return [[self.center[0], self.center[1]],[self.center[0], self.center[1]]]
            return None
        cos2Theta0 = self.S1 / S0
        sin2Theta0 = self.S2 / S0
        sinTheta0 = math.sqrt(0.5 * (1. - cos2Theta0))
        cosTheta0 = 0.5 * sin2Theta0 / sinTheta0
        scale = rescaling_factor * S0
        xBegin = self.center[1] - scale * cosTheta0
        xEnd = self.center[1] + scale * cosTheta0
        yBegin = self.center[0] - scale * sinTheta0
        yEnd = self.center[0] + scale * sinTheta0
        pts = []
        pts.append([yBegin, xBegin])
        pts.append([yEnd, xEnd])
        return pts

    #
    #     /**
    #      * Converts a nematic to a MyLine2D
    #      *
    #      * @param rescaling_factor
    #      * @return a vectorial line object corresponding to the nematic
    #      */
    #     public MyLine2D.Double toMyLine2D(double rescaling_factor) {
    #         Point2D.Double[] base_n_tip = getBeginAndEnd(rescaling_factor);
    #         return new MyLine2D.Double(base_n_tip[0], base_n_tip[1]);
    #     }
    #
    #     public void draw(Graphics2D g2d) {
    #         toMyLine2D(1).drawAndFill(g2d);
    #     }
    #
    #     public boolean isNaN() {
    #         return (Double.isNaN(S1) || Double.isNaN(S2));
    #     }
    #

    # drawing in python over numpy array https://stackoverflow.com/questions/28647383/numpy-compatible-image-drawing-library --> maybe try the other libs such as wand because may be interesting

    def toLine2D(self, rescaling_factor=1.):
        base_n_tip = None
        try:
             base_n_tip =self.getBeginAndEnd(rescaling_factor)
        except:
            # traceback.print_exc()
            # print('error nematic could not be converted to Line2D')
            return None

        # print('self.S1, self.S2',self.S1, self.S2)
        if base_n_tip is None:
            return None
        # line(int(math.ceil(base_n_tip[0][0])), int(math.ceil(base_n_tip[0][1])),
        #      int(math.ceil(base_n_tip[0][1])), int(math.ceil(base_n_tip[1][1])))
        # print(base_n_tip[0][0], base_n_tip[0][1],base_n_tip[1][1], base_n_tip[1][1])
        return Line2D(base_n_tip[0][1], base_n_tip[0][0], base_n_tip[1][1], base_n_tip[1][0]) # KEEP NB BE careful --> first x then y and not the opposite --> maybe change this because dangerous in python where everything is inverted but then need change everythng


    # nb please set the rescaling factor or set it here
    def draw(self, img, color=0xFFFF00, stroke=1, rescaling_factor=None):
        if color is None:
            color = 0xFFFF00
        if rescaling_factor is None:
            base_n_tip = self.getBeginAndEnd(self.rescaling_factor)
        else:
            base_n_tip = self.getBeginAndEnd(rescaling_factor)
        if base_n_tip is None:
            # null nematic --> we don't plot it
            return
        if isinstance(img, np.ndarray):
            # draw the nematic on the image

            # new
            # MyLine2D.Double(base_n_tip[0], base_n_tip[1]);
            rr, cc = line(int(math.ceil(base_n_tip[0][0])), int(math.ceil(base_n_tip[0][1])),
                          int(math.ceil(base_n_tip[1][0])), int(math.ceil(base_n_tip[1][1])))


            # if out of bonds --> errors --> really sucks --> see how I can fix that !!!
            img[rr, cc] = color
        elif isinstance(img, ImageDraw):
            if isinstance(stroke, str):
                stroke = float(stroke)
            if stroke is None or stroke <= 0:
                stroke = 1
            if stroke < 1:
                stroke = int(1)
            if isinstance(stroke, float):
                stroke = int(round(stroke))
            if isinstance(color, str):
                if color.startswith('#'):
                    # parse it or maybe best is to convert the entire column rather than do things here
                    color = color.replace('#', '')
                    color = int(color, 16)

            color = ((color >> 16) & 255, (color >> 8) & 255, (color & 255))
            # print(color)

            # print(base_n_tip)
            # is there a bug ???? --> yes the coords must be given as x and y --> opposite of numpy!!!!
            img.line((base_n_tip[0][1], base_n_tip[0][0], base_n_tip[1][1], base_n_tip[1][0]), fill=color, width=stroke)
            # img.line((350, 200, 450, 100), fill=(255, 255, 0), width=10)

        #         return new MyLine2D.Double(base_n_tip[0], base_n_tip[1]);

    # return the components of the unit Nematic
    def getUnitNematicComponents(self):
        S0 = self.getMagnitude()
        return [self.S1 / S0, self.S2 / S0]

    #
    #     /**
    #      *
    #      * @return the unit nematic
    #      */
    # return the unit nematic
    def getUnitNemat(self):
        unit0 = self.getUnitNematicComponents()
        return Nematic(S1=unit0[0], S2=unit0[1], center=self.center)

    def __str__(self):
        return str(self.S1) + " " + str(self.S2) + " " + str(self.center) + " " + str(self.getS0()) + " " + str(
            self.getAngleInRadians()) + " " + str(self.getBeginAndEnd()[0]) + " " + str(self.getBeginAndEnd()[1])
        # return (S1 + " " + S2 + " " + center + " " + this.getS0() + " " + this.getAngleInRadians() + " " +
        #         this.getBeginAndEnd(1)[0] + " " + this.getBeginAndEnd(1)[1]);





if __name__ == "__main__":

    # props = measure.regionprops(
    #     labels, image, extra_properties=[image_height]
    # )

    if True:
        base = (301.77769935043347,        21.280930792689446)
        tip = (230.48494928416977,        715.8784747926309)

        from epyseg.ta.measurements.TAmeasures import distance_between_points

        nematic = Nematic(base=base, tip=tip)

        print(nematic)

        print(nematic.getBeginAndEnd())

        # so if I wanna scale it -> I can easily do that
        print('rescaled', nematic.getBeginAndEnd(rescaling_factor=0.75))

        dist_reduced = distance_between_points(*nematic.getBeginAndEnd(rescaling_factor=0.75))
        orig_dist = distance_between_points(base, tip)
        print(dist_reduced,orig_dist, dist_reduced/orig_dist) # --> all is perfect --> I can really use that

        import sys
        sys.exit(0)

    if True:
        handcorrection = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/handCorrection.tif')
        handcorrection = label(handcorrection, connectivity=1, background=255)
        rps = regionprops(handcorrection) # , extra_properties=[compute_stretch_nematic_regionprops_extra] --> marche pas car il m'envoir juste un mask et ça ne m'interesse pas...



        first_cell = rps[0]

        # print(rps[0].compute_stretch_nematic_regionprops_extra)
        # print(first_cell.coords.shape)
        # put it in a coordinate list --> TODO --> maybe also allow to pass in the array
        # convert the coords to a numpy array and loop over it
        # print(type(first_cell))
        print(compute_stretch_nematic(first_cell.coords))
        print(compute_stretch_nematic(first_cell))



        sys.exit(0)

    if True:
        Q1 = -42.17042288690961
        Q2 = -58.92645735167784

        xCenter = 449.56
        yCenter = 344.48
        test3 = Nematic(center=[yCenter, xCenter], S1=Q1, S2=Q2)

        print(test3)

        print(test3.get_angle2())
        print(test3.getAngleInRadians())
        # in TA --> -42.17042288690961 -58.92645735167784 Point2D.Double[449.56, 344.48] 72.46152042622275 2.0456200121341603 Point2D.Double[482.68807708785437, 280.03465531986757] Point2D.Double[416.43192291214564, 408.92534468013247]

    if False:
        xCenter = 256
        yCenter = 256
        Q1 = 215.43
        Q2 = -0.4503167411  # //-0.4503167411
        #             //double Q1 = -79.9;
        #             //double Q2 = -75.09;
        #             test3 = Nematic(center=[xCenter, yCenter], S1=Q1, S2=Q2)
        test3 = Nematic(center=[xCenter, yCenter], S0=Q1,
                        angle=Q2)  # seems to work as in TA --> just need invert center coords to fit
        print(test3)

    # TA  133.80650714020177 -168.83691405299712 Point2D.Double[256.0, 256.0] 215.43 2.691275912489793 Point2D.Double[449.95362942365074, 162.2339371979757] Point2D.Double[62.04637057634926, 349.7660628020243]
    # PyTA 133.80650714020177 -168.83691405299712 [256, 256] 215.43 2.691275912489793 [162.2339371979757, 449.95362942365074] [349.7660628020243, 62.04637057634926]

    if False:
        tmp = Nematic(center=[20, 10], S0=10.44030650891055,
                      angle=0.1457283972389337);  # //new Nematic(new Point2D.Double(10, 20), new Point2D.Double(20.329643389031034, 21.516069739638404)) #//new Nematic(-10, 3, new Point2D.Double(10, 20));//new Nematic(-1020, -10000, new Point2D.Double(25, 25));
        print(tmp)
        # //      print(tmp.getAngleInRadians() , tmp.getAngle2() + " " + tmp.getAngleTest());
        # //        tmp.getAngleTest();

        test = Nematic(S1=-10, S2=3, center=[20, 10])
        print(test)
        # //        System.out.println(test);

        # //--> ca marche il y a une tte petite reounding error
        # test2 = Nematic([10, 20], [20.329643389031034, 21.516069739638404]) --> not implemented --> center and any extremity...
        # //        System.out.println(test2);

        test3 = Nematic(center=[20, 10], S0=10.44030650891055, angle=0.1457283972389337)
        # //        System.out.println(test3);

        unit_nemat = test3.getUnitNematicComponents()
        print(unit_nemat)
        print(math.sqrt(unit_nemat[1] * unit_nemat[1] + unit_nemat[0] * unit_nemat[0]))

        print(test3.getUnitNemat().getMagnitude())

    # nb all seems fine now

# 215.43
# 154.19875129088464
# Point2D.Double[449.95362942365074, 162.2339371979757] Point2D.Double[62.04637057634926, 349.7660628020243]
# 69.64703070271749
# 54.08721586034971
# 15.559814842367771
# -15.559814842367771
# Point2D.Double[0.9578262852211513, 0.28734788556634566]
# 1.0
# 1.0

#             System.out.println(test3.getMagnitude());
#             System.out.println(Math.toDegrees(test3.getAngleInRadians()));
#             BufferedImage test = new BufferedImage(512, 512, BufferedImage.TYPE_INT_RGB);
#             Graphics2D g2d = test.createGraphics();
#             test3.toMyLine2D(1).drawAndFill(g2d);
#
#             System.out.println(test3.getBeginAndEnd()[0] + " " + test3.getBeginAndEnd()[1]);
#
#             MyLine2D.Double line = new MyLine2D.Double(xCenter - Q1 / 2, yCenter - Q2 / 2, xCenter + Q1, yCenter + Q2);
#             line.setDrawColor(0xFF0000);
#             line.drawAndFill(g2d);
#             g2d.dispose();
#             Saver.popJ(test);
#             try {
#                 Thread.sleep(3000);
#             } catch (InterruptedException ex) {
#                 //  Logger.getLogger(Nematic.class.getName()).log(Level.SEVERE, null, ex);
#             }
#             //System.exit(0);
#         }
#
#         System.out.println(Math.toDegrees(1.21557));
#         System.out.println(Math.toDegrees(0.944));
#
#         System.out.println(Math.toDegrees(1.21557 - 0.944));
#         System.out.println(Math.toDegrees(-1.21557 + 0.944));//System.out.println(Math.toDegrees(0.26777283585356804));
#
# //        ArrayList<String> list = new LoadListeToArrayList().apply("/list.lst");
#         //long start_time = System.currentTimeMillis();
# //        int x1 = 2;
# //        int x2 = 3;
# //        double val = (x1+x2)/2.;
# //        System.out.println(val);
# //        if (true)
# //        return;
#         //--> tjrs le deuxieme
#         Nematic tmp = new Nematic(new Point2D.Double(10, 20), 10.44030650891055, 0.1457283972389337);//new Nematic(new Point2D.Double(10, 20), new Point2D.Double(20.329643389031034, 21.516069739638404));//new Nematic(-10, 3, new Point2D.Double(10, 20));//new Nematic(-1020, -10000, new Point2D.Double(25, 25));
# //        System.out.println(tmp.getAngleInRadians() + " " + tmp.getAngle2() + " " + tmp.getAngleTest());
# //        tmp.getAngleTest();
#
#         Nematic test = new Nematic(-10, 3, new Point2D.Double(10, 20));
# //        System.out.println(test);
#
#         //--> ca marche il y a une tte petite reounding error
#         Nematic test2 = new Nematic(new Point2D.Double(10, 20), new Point2D.Double(20.329643389031034, 21.516069739638404));
# //        System.out.println(test2);
#
#         Nematic test3 = new Nematic(new Point2D.Double(10, 20), 10.44030650891055, 0.1457283972389337);
# //        System.out.println(test3);
#
#         Point2D.Double unit_nemat = test3.getUnitNematicComponents();
#         System.out.println(unit_nemat);
#         System.out.println(Math.sqrt(unit_nemat.x * unit_nemat.x + unit_nemat.y * unit_nemat.y));
#         //--> is ok
#
#         System.out.println(test3.getUnitNemat().getMagnitude());
#
#         //bonne idee il faut tester les nematics
#         //et leur normalisation
# //        test.apply(list);
#         //System.out.println("ellapsed time --> " + (System.currentTimeMillis() - start_time) / 1000.0 + "s");
#         System.exit(0);
#     }
# }



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy.QtCore import QTimer


class Blinker:

    def __init__(self):
        pass

    def blink(self, component):
        if isinstance(component, list):
            for comp in component:
                comp.setStyleSheet("background-color: red;")
        else:
            component.setStyleSheet("background-color: red;")

        self.current_timer = QTimer()
        self.current_timer.timeout.connect(lambda: self.reset_stylesheet(component))
        self.current_timer.setSingleShot(True)
        self.current_timer.start(3000)

    def reset_stylesheet(self, component):
        if isinstance(component, list):
            for comp in component:
                comp.setStyleSheet('')
        else:
            component.setStyleSheet('')

if __name__ == '__main__':
    blinker = Blinker()
    # blinker.blink()



from epyseg.ta.GUI.paint2 import Createpaintwidget
from epyseg.ta.GUI.scrollablepaint import scrollable_paint

# this is the perfect TA like drawing area --> just need replace it
class tascrollablepaint(scrollable_paint):
    def __init__(self, overriding_paint_widget=None):
        if overriding_paint_widget is None:
            class overriding_apply(Createpaintwidget):
                # all seems ok now and functions as in TA
                # just do the shift enter to get rid of small cells

                # avec ça ça marche 100% à la TA ... --> cool --> maybe make it a TA drawing pad
                def apply(self):
                    self.apply_drawing(minimal_cell_size=0)

                def shift_apply(self):
                    # MEGA TODO IMPLEMENT SIZE within this stuff!!!
                    self.apply_drawing(minimal_cell_size=10)

                def ctrl_m_apply(self):
                    self.manually_reseeded_wshed() # how can I pass the channel to the stuff ???
                    # channel = self.get_selected_channel()

                def m_apply(self):
                    self.maskVisible = not self.maskVisible
                    self.update()

                def save(self):
                    self.save_mask()

            super().__init__(custom_paint_panel=overriding_apply())
        else:
            super().__init__(custom_paint_panel=overriding_paint_widget)


if __name__ == '__main__':
    import os
    from epyseg.settings.global_settings import set_UI  # set the UI to be used py qtpy
    set_UI()
    import sys
    from qtpy.QtWidgets import QApplication
    if False:
        app = QApplication(sys.argv)
        qimage = toQimage(Img('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/neo_mask.tif'))
        # seg fault here
        print('before seg fault')
        icon = QIcon(QPixmap.fromImage(qimage))  # b
        print(qimage, icon)
        sys.exit(0)


    # TODO add a main method so it can be called directly
    # maybe just show a canvas and give it interesting props --> TODO --> really need fix that too!!!

    app = QApplication(sys.argv)

    w = tascrollablepaint()  # ça marche --> permet de mettre des paint panels avec des proprietes particulieres --> assez facile en fait
    # w.set_image('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/Optimized_projection_018.png')
    w.set_image('/E/Sample_images/sample_images_different_sizes_and_bit_depth_and_nb_of_channels/210219.lif_t000.tif')
    w.set_mask('/E/Sample_images/sample_images_different_sizes_and_bit_depth_and_nb_of_channels/210219.lif_t000/handCorrection.tif')
    # w.set_image('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/neo_mask.tif') # nb if alone it is causing a crash --> why # --> incomprehensible why bugs sometimes
    # w.set_mask('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/neo_mask.tif')


    w.show()
    sys.exit(app.exec_())





# modified from https://stackoverflow.com/questions/24469662/how-to-redirect-logger-output-into-pyqt-text-widget
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import sys
from qtpy import QtCore
import logging

class QtHandler(logging.Handler):
    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        record = self.format(record)
        if record:
            if record.startswith('ERROR') or record.startswith('CRITICAL') or record.startswith('WARNING'):
                XStream.stderr().write('%s' % record)
            else:
                XStream.stdout().write('%s' % record)


class XStream(QtCore.QObject):
    _stdout = None
    _stderr = None
    messageWritten = QtCore.Signal(str)

    def flush(self):
        pass

    def fileno(self):
        return -1

    def write(self, msg):
        if (not self.signalsBlocked()):
            self.messageWritten.emit(msg)

    @staticmethod
    def stdout():
        if (not XStream._stdout):
            XStream._stdout = XStream()
            sys.stdout = XStream._stdout
        return XStream._stdout

    @staticmethod
    def stderr():
        if (not XStream._stderr):
            XStream._stderr = XStream()
            sys.stderr = XStream._stderr
        return XStream._stderr



# I guess there is a big bug in the scales of the components --> hack this

# faire un draw_at_scale and merge draw fill etc by just putting an option --> much simpler and that way you can keep original size which is the best idea
# set to width doit juste calculer le scaling factor in fact.... --> TODO do that and same for all objects --> no cloning and all will be simpler

# check all the things that need be packed top left/top right/bottom left/bottom right
# letter should always be packed first and top left
# make insets, etc more flexible
# should I remove the letter of insets ???
# the rest can be any position and any nb of instances of text/scale bar or insets
# see how to edit that easily
# TODO
# objects can be scale bars and or insets and or text labels as many as needed --> pack them in x or y and align them

# https://docs.python.org/2/library/operator.html
# maths fig in inkscape --> cool using latex https://castel.dev/post/lecture-notes-2/

# TODO may also contain svg or graphs or ???
# TODO handle extra labels for images directly and also for rows or cols --> think how to do that but must be doable

# TODO Add the crops --> see how though and see how to warn when width or height reaches 0...
# should I offer max projections too ??? maybe or not --> see how to do that...

# the only advantage I see for rect2D vs qrectf is that they allow for rotation ????
# see how to best handle that ???
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import numpy as np
import matplotlib.pyplot as plt
from qtpy import QtGui
from qtpy.QtGui import QPainter, QColor
from qtpy.QtSvg import QSvgGenerator
import os
from epyseg.draw.shapes.line2d import Line2D
from epyseg.draw.shapes.point2d import Point2D
from epyseg.figure.alignment import alignRight, alignLeft, alignTop, alignBottom, alignCenterH, alignCenterV, packY, \
    packX, packYreverse
from epyseg.draw.shapes.rect2d import Rect2D
from epyseg.draw.shapes.scalebar import ScaleBar
from epyseg.draw.shapes.txt2d import TAText2D
from epyseg.figure.fig_tools import preview
# from epyseg.figure import fig_tools
from epyseg.img import Img, toQimage
from qtpy.QtCore import QRectF, QPointF, QSize, QRect
# from sympy import nsolve, exp, Symbol
# logger
from epyseg.tools.logger import TA_logger

logger = TA_logger()

class Image2D(Rect2D):
    TOP_LEFT = 0
    TOP_RIGHT = 1
    BOTTOM_LEFT = 2
    BOTTOM_RIGHT = 3
    CENTERED = 4

    def __init__(self, *args, x=None, y=None, width=None, height=None, data=None, dimensions=None, opacity=1., **kwargs):
        self.isSet = False
        self.scale = 1
        self.translation = QPointF()

        # crops
        self.__crop_left = 0
        self.__crop_right = 0
        self.__crop_top = 0
        self.__crop_bottom = 0
        self.img = None
        self.annotation = []  # should contain the objects for annotating imaging --> shapes and texts
        self.letter = None  # when objects are swapped need change the letter
        self.top_left_objects = []
        self.top_right_objects = []
        self.bottom_right_objects = []
        self.bottom_left_objects = []
        self.centered_objects = []

        # if the image is inserted as an inset then draw it as a fraction of parent width
        # inset parameters
        self.fraction_of_parent_image_width_if_image_is_inset = 0.25
        self.border_size = None  # no border by default
        self.border_color = 0xFFFFFF  # white border by default

        if args:
            if len(args) == 1:
                if isinstance(args[0], str):
                    self.filename = args[0]
                elif isinstance(args[0], Img):
                    self.filename = None
                    self.img = args[0]
                    self.qimage =  toQimage(self.img)
                    if x is None:
                        x = 0
                    if y is None:
                        y = 0
                    try:
                        super(Image2D, self).__init__(x, y, self.img.get_width(), self.img.get_height())
                    except:
                        super(Image2D, self).__init__(x, y, self.img.shape[1], self.img.shape[0])
                    self.isSet = True
        else:
            self.filename = None

        if x is None and y is None and width is not None and height is not None:
            super(Image2D, self).__init__(0, 0, width, height)
            self.isSet = True
        elif x is None and y is None and width is None and height is None and self.filename is not None:
            # print('in 0')
            try:
                self.img = Img(self.filename)
            except:
                logger.error('could not load image '+str(self.filename))
                return
            self.qimage = toQimage(self.img)
            width = self.img.get_width()
            height = self.img.get_height()
            super(Image2D, self).__init__(0, 0, width, height)
            self.isSet = True
        elif x is not None and y is not None and width is not None and height is not None and self.img is None:
            self.img = None
            super(Image2D, self).__init__(x, y, width, height)
            self.isSet = True
        elif data is None:
            if self.filename is not None:
                self.img = Img(self.filename)
                self.qimage =  toQimage(self.img)
                if x is None:
                    x = 0
                if y is None:
                    y = 0
                super(Image2D, self).__init__(x, y, self.img.get_width(), self.img.get_height())
                self.isSet = True
        elif data is not None:
            self.img = Img(data,
                           dimensions=dimensions)  # need width and height so cannot really be only a numpy stuff --> cause no width or height by default --> or need tags such as image type for dimensions
            self.qimage =  toQimage(self.img)
            # need Image dimensions id data is not of type IMG --> could check that
            if x is None:
                x = 0
            if y is None:
                y = 0
            super(Image2D, self).__init__(x, y, self.img.get_width(), self.img.get_height())
            self.isSet = True
        self.opacity = opacity

    # @return the block incompressible width
    def getIncompressibleWidth(self):
        extra_space = 0  # can add some if boxes around to add text
        return extra_space

    # @return the block incompressible height
    def getIncompressibleHeight(self):
        extra_space = 0  # can add some if boxes around to add text
        # do not even do that --> just ignore things
        # add box around ???
        return extra_space

    def add_object(self, object, position):
        if isinstance(object, list):
            for obj in object:
                self.add_object(obj, position=position)
            return
        if position == Image2D.TOP_LEFT:
            self.top_left_objects.append(object)
        elif position == Image2D.BOTTOM_RIGHT:
            self.bottom_right_objects.append(object)
        elif position == Image2D.BOTTOM_LEFT:
            self.bottom_left_objects.append(object)
        elif position == Image2D.CENTERED:
            self.centered_objects.append(object)
        else:
            self.top_right_objects.append(object)

    # TODO --> check if contains it
    def remove_object(self, object, position):
        if position == Image2D.TOP_LEFT:
            self.top_left_objects.remove(object)
        elif position == Image2D.BOTTOM_RIGHT:
            self.bottom_right_objects.remove(object)
        elif position == Image2D.BOTTOM_LEFT:
            self.bottom_left_objects.remove(object)
        elif position == Image2D.CENTERED:
            self.centered_objects.remove(object)
        else:
            self.top_right_objects.remove(object)

    def remove_all_objects(self, position):
        if position == Image2D.TOP_LEFT:
            del self.top_left_objects
            self.top_left_objects = []
        elif position == Image2D.BOTTOM_RIGHT:
            del self.bottom_right_objects
            self.bottom_right_objects = []
        elif position == Image2D.BOTTOM_LEFT:
            del self.bottom_left_objects
            self.bottom_left_objects = []
        elif position == Image2D.CENTERED:
            del self.centered_objects
            self.centered_objects = []
        else:
            del self.top_right_objects
            self.top_right_objects = []

    def setLettering(self, letter):
        if isinstance(letter, TAText2D):
            self.letter = letter
        elif isinstance(letter, str):
            if letter.strip() == '':
                self.letter = None
            else:
                self.letter = TAText2D(letter)

    # def getRect2D(self):
    #     # self.__class__ = Rect2D
    #     # return super()
    #     # TODO ideally I'd like to get the Rect2D parent but I should think what the best way is to get it...
    #     return self

    def draw(self, painter, draw=True):
        if draw:
            painter.save()
            painter.setOpacity(self.opacity)
            # painter.setClipRect(self)  # only draw in self --> very useful for inset borders # pb clip rect does not work for svg --> remove for now users can add it manually if desired or I can add it if people really want it and then I should draw relevant lines or shifted rects --> do that later
        # prevents drawing outside from the image
            rect_to_plot = self.boundingRect(scaled=True) #scaled=True #self.adjusted(self.__crop_left, self.__crop_top, self.__crop_right, self.__crop_bottom) # need remove the crops with that
            # self.scale = 1
            # if self.scale is not None and self.scale != 1:
            # #     # TODO KEEP THE ORDER THIS MUST BE DONE THIS WAY OR IT WILL GENERATE PLENTY OF BUGS...
            #     new_width = rect_to_plot.width() * self.scale
            #     new_height = rect_to_plot.height() * self.scale
            # #     # print(rect_to_plot.width(), rect_to_plot.height())  # here ok
            # #     # setX changes width --> why is that
            # #
            # #     # TODO BE EXTREMELY CAREFUL AS SETX AND SETY CAN CHANGE WIDTH AND HEIGHT --> ALWAYS TAKE SIZE BEFORE OTHERWISE THERE WILL BE A PB AND ALWAYS RESET THE SIZE WHEN SETX IS CALLED!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # #     # Sets the left edge of the rectangle to the given x coordinate. May change the width, but will never change the right edge of the rectangle. --> NO CLUE WHY SHOULD CHANGE WIDTH THOUGH BUT BE CAREFUL!!!
            # #     rect_to_plot.setX(rect_to_plot.x() * self.scale)
            # #     rect_to_plot.setY(rect_to_plot.y() * self.scale)
            # #     # maybe to avoid bugs I should use translate instead rather that set x but ok anyways
            # #     # print(rect_to_plot.width(), rect_to_plot.height())# bug here --> too big
            # #
            # #     # print(new_height, new_height, self.width(), self.scale, self.scale* self.width())
            #     rect_to_plot.setWidth(new_width)
            #     rect_to_plot.setHeight(new_height)

            if self.img is not None:
                x = 0
                y = 0
                try:
                    w = self.img.get_width()
                    h = self.img.get_height()
                except:
                    w = self.img.shape[1]
                    h = self.img.shape[0]

                if self.__crop_top is not None:
                    y = self.__crop_top
                    h -= self.__crop_top
                if self.__crop_left is not None:
                    x = self.__crop_left
                    w -= self.__crop_left
                if self.__crop_right is not None:
                    w -= self.__crop_right
                if self.__crop_bottom is not None:
                    h -= self.__crop_bottom
                # pb here --> see how to really crop
                qsource = QRectF(x, y, w, h)
                painter.drawImage(rect_to_plot, self.qimage, qsource)  # , flags=QtCore.Qt.AutoColor
            else:
                painter.drawRect(rect_to_plot)



            # letter is good
            extra_space = 3

            # draw annotations first
            if self.annotation is not None and self.annotation:
                # need clone the object then set its P1 with respect to position or need a trick to keep original ref and have an updated one just for display but then need renew it all the time --> see how I can do that...
                # maybe clone is not smart as it duplicates resources without a need for it
                # but then need clone the original rect and draw with respect to that
                # and I indeed need scale the shape --> TODO too
                # indeed thanks to cloning I always preserve original info --> not bad

                # annot position is good
                # TODO see how to do that cause not so easy --> think carefully and take inspiration from EZF and improve it
                for annot in self.annotation:
                    # always empty --> why is that
                    # print('init',annot.get_P1())
                    # always assume everything is done at 0,0 then do translation
                    # annot.set_P1(self.get_P1().x() + annot.get_P1().x(),                    self.get_P1().y() + annot.get_P1().y())  # always relative to the parent image
                    # annot.set_P1(self.get_P1())  # always relative to the parent image
                    # print(annot.get_P1())

                    # print('init', self.get_P1(), 'scale', self.get_scale())
                    annot.set_to_translation(rect_to_plot.topLeft())

                    annot.set_to_scale(self.scale)  # will fuck the stuff but ok for a test
                    # print('scaled',annot.get_P1())
                    annot.draw(painter=painter)
                    # print('tranbs', annot.translation)




            # and indeed I need also to take crop into account in order not to misposition things...

            if self.letter is not None:
                self.letter.set_P1(rect_to_plot.topLeft().x() + extra_space, rect_to_plot.topLeft().y() + extra_space)

            # then draw text and insets --> on top of annotations
            # TODO need align insets differently than others and need align its bounding box also differently --> TODO but almost there
            if len(self.top_right_objects) != 0 or len(self.top_left_objects) != 0 or len(
                    self.bottom_left_objects) != 0 or len(self.bottom_right_objects) != 0 or len(
                    self.centered_objects) != 0:
                # align a scale bar to various positions
                # maybe if there is a letter first point should be place below stuff
                # top_left = Point2D(self.get_P1())
                top_left_shifted = Point2D(rect_to_plot.topLeft())
                # top_left_shifted.setX(top_left_shifted.x() )# + extra_space
                # top_left_shifted.setY(top_left_shifted.y() )#+ extra_space

                # print('before', top_left)
                # if self.letter is not None:
                #     packY(extra_space, self.letter, top_left_shifted)
                # print('after', top_left)

                # insets should be aligned to unshifted values
                # whereas texts should be aligned to shifted ones
                # what if I try all unshifted
                # cause in a way it's simpler

                # top_right = Point2D(self.get_P1())
                top_right_shifted = Point2D(rect_to_plot.topLeft())
                top_right_shifted.setX(top_right_shifted.x() + rect_to_plot.width())#- extra_space
                top_right_shifted.setY(top_right_shifted.y() )#+ extra_space

                # bottom_left = Point2D(self.get_P1())
                bottom_left_shifted = Point2D(rect_to_plot.topLeft())
                bottom_left_shifted.setX(bottom_left_shifted.x() ) #+ extra_space
                bottom_left_shifted.setY(
                    bottom_left_shifted.y() + rect_to_plot.height() )#- extra_space  # should align right then pack on top of that --> may need a direction in packing--> TODO

                bottom_right = Point2D(rect_to_plot.topLeft())
                bottom_right_shifted = Point2D(rect_to_plot.topLeft())
                bottom_right_shifted.setX(bottom_right_shifted.x() + rect_to_plot.width())# - extra_space
                bottom_right_shifted.setY(bottom_right_shifted.y() + rect_to_plot.height())#- extra_space

                center = Point2D(rect_to_plot.topLeft())
                center.setX(center.x() + rect_to_plot.width() / 2)
                center.setY(center.y() + rect_to_plot.height() / 2)

                if len(self.top_left_objects) != 0:
                    # change inset size first
                    for obj in self.top_left_objects:
                        if isinstance(obj, Image2D):
                            obj.setToWidth(rect_to_plot.width() * obj.fraction_of_parent_image_width_if_image_is_inset)

                    # if letter exists align with respect to it

                    alignTop(top_left_shifted, *self.top_left_objects)
                    alignLeft(top_left_shifted, *self.top_left_objects)

                    if self.letter is not None:
                        # packY(extra_space, self.letter, top_left_shifted)
                        top_left_shifted = self.letter

                    # in fact images really need be aligned left of the image but the others need be aligned with the letter that has an extra space --> TODO --> change some day

                    packY(extra_space, top_left_shifted, *self.top_left_objects)

                    # all images need be shifted back??? to be aligned left

                    for obj in self.top_left_objects:
                        # for drawing of inset borders
                        # if isinstance(obj, Image2D):
                        #     # make it draw a border and align it
                        #     # painter.save()
                        #     img_bounds = Rect2D(obj)
                        #     img_bounds.stroke = 3
                        #     # img_bounds.translate(-img_bounds.stroke / 2, -img_bounds.stroke / 2)
                        #     img_bounds.color = 0xFFFF00
                        #     img_bounds.fill_color = 0xFFFF00
                        #     img_bounds.draw(painter=painter)
                        #     # print(img_bounds)
                        #     # painter.restore()
                        obj.draw(painter=painter)

                if len(self.top_right_objects) != 0:
                    # change inset size first
                    for obj in self.top_right_objects:
                        if isinstance(obj, Image2D):
                            obj.setToWidth(rect_to_plot.width() * obj.fraction_of_parent_image_width_if_image_is_inset)
                    alignRight(top_right_shifted, *self.top_right_objects)
                    alignTop(top_right_shifted, *self.top_right_objects)
                    packY(extra_space, top_right_shifted, *self.top_right_objects)
                    for obj in self.top_right_objects:
                        # # for drawing of inset borders
                        # if isinstance(obj, Image2D):
                        #     # make it draw a border and align it
                        #     # painter.save()
                        #     img_bounds = Rect2D(obj)
                        #     img_bounds.stroke = 3
                        #     # img_bounds.translate(img_bounds.stroke / 2, -img_bounds.stroke / 2)
                        #     img_bounds.color = 0xFFFF00
                        #     img_bounds.fill_color = 0xFFFF00
                        #     img_bounds.draw(painter=painter)
                        #     # print(img_bounds)
                        #     # painter.restore()
                        obj.draw(painter=painter)

                if len(self.bottom_right_objects) != 0:
                    # change inset size first
                    for obj in self.bottom_right_objects:
                        if isinstance(obj, Image2D):
                            obj.setToWidth(rect_to_plot.width() * obj.fraction_of_parent_image_width_if_image_is_inset)
                    alignRight(bottom_right_shifted, *self.bottom_right_objects)
                    alignBottom(bottom_right_shifted, *self.bottom_right_objects)
                    packYreverse(extra_space, bottom_right_shifted, *self.bottom_right_objects)
                    # packY(3, top_right, *self.top_right_objects) # I do need to invert packing order
                    for obj in self.bottom_right_objects:
                        # # for drawing of inset borders
                        # if isinstance(obj, Image2D):
                        #     # make it draw a border and align it
                        #     # painter.save()
                        #     img_bounds = Rect2D(obj)
                        #     img_bounds.stroke = 3
                        #     # img_bounds.translate(-img_bounds.stroke / 2, img_bounds.stroke / 2)
                        #     # should I clip it to the image size --> maybe it's the best
                        #     img_bounds.color = 0xFFFF00
                        #     img_bounds.fill_color = 0xFFFF00
                        #     img_bounds.draw(painter=painter)
                        #     # print(img_bounds)
                        #     # painter.restore()
                        obj.draw(painter=painter)

                if len(self.bottom_left_objects) != 0:
                    # change inset size first
                    for obj in self.bottom_left_objects:
                        if isinstance(obj, Image2D):
                            obj.setToWidth(rect_to_plot.width() * obj.fraction_of_parent_image_width_if_image_is_inset)
                    alignLeft(bottom_left_shifted, *self.bottom_left_objects)
                    alignBottom(bottom_left_shifted, *self.bottom_left_objects)
                    packYreverse(extra_space, bottom_left_shifted, *self.bottom_left_objects)
                    for obj in self.bottom_left_objects:
                        # # for drawing of inset borders
                        # if isinstance(obj, Image2D):
                        #     # make it draw a border and align it
                        #     # painter.save()
                        #     img_bounds = Rect2D(obj)
                        #     img_bounds.stroke = 3
                        #     # img_bounds.translate(-img_bounds.stroke/2, img_bounds.stroke/2)
                        #     img_bounds.color = 0xFFFF00
                        #     img_bounds.fill_color = 0xFFFF00
                        #     img_bounds.draw(painter=painter)
                        #     # print(img_bounds)
                        #     # painter.restore()
                        obj.draw(painter=painter)

                if len(self.centered_objects) != 0:
                    # change inset size first
                    for obj in self.centered_objects:
                        if isinstance(obj, Image2D):
                            obj.setToWidth(rect_to_plot.width() * obj.fraction_of_parent_image_width_if_image_is_inset)
                    alignCenterH(center, *self.centered_objects)
                    alignCenterV(center, *self.centered_objects)
                    for obj in self.centered_objects:
                        # # for drawing of inset borders
                        # if isinstance(obj, Image2D):
                        #     # make it draw a border and align it
                        #     # painter.save()
                        #     img_bounds = Rect2D(obj)
                        #     img_bounds.stroke = 3
                        #     img_bounds.color = 0xFFFF00
                        #     img_bounds.fill_color = 0xFFFF00
                        #     img_bounds.draw(painter=painter)
                        #     # print(img_bounds)
                        #     # painter.restore()
                        obj.draw(painter=painter)

            # then need to draw the letter at last so that it is always on top
            if self.letter is not None:
                self.letter.draw(painter)

            painter.restore()
                # # TOP left 2
                # scale_bar = ScaleBar(30, '<font color="#FF00FF">10µm</font>')
                # scale_bar.set_scale(self.get_scale())
                # # scale_bar.set_P1(self.get_P1().x()+extra_space, self.get_P1().y()+extra_space)
                # scale_bar.set_P1(self.get_P1())
                # alignLeft(top_left, scale_bar)
                # alignTop(top_left, scale_bar)
                # # scale_bar.set_P1(scale_bar.get_P1().x()-extra_space, scale_bar.get_P1().x()+extra_space)
                # scale_bar.drawAndFill(painter=painter)

                # TOP right 2
                # scale_bar = ScaleBar(30, '<font color="#FF00FF">10µm</font>')
                # scale_bar.set_scale(self.get_scale())
                # # scale_bar.set_P1(self.get_P1().x()+extra_space, self.get_P1().y()+extra_space)
                # scale_bar.set_P1(self.get_P1())
                # alignRight(top_right, scale_bar)
                # alignTop(top_right, scale_bar)
                # # scale_bar.set_P1(scale_bar.get_P1().x()-extra_space, scale_bar.get_P1().x()+extra_space)
                # scale_bar.drawAndFill(painter=painter)

                # bottom left 2

                # # big bug in scale --> the size of the stuff isn't respected
                # # 288 is the size of image 0
                # scale_bar = ScaleBar(288, '<font color="#FF00FF">10µm</font>')
                # scale_bar.set_scale(self.get_scale())
                # # scale_bar.set_P1(self.get_P1().x()+extra_space, self.get_P1().y()+extra_space)
                # scale_bar.set_P1(self.get_P1())
                # alignLeft(bottom_left, scale_bar)
                # alignBottom(bottom_left, scale_bar)
                # # scale_bar.set_P1(scale_bar.get_P1().x()-extra_space, scale_bar.get_P1().x()+extra_space)
                # scale_bar.drawAndFill(painter=painter)

                # # bottom right 2
                # scale_bar = ScaleBar(30, '<font color="#FF00FF">10µm</font>')
                # scale_bar.set_scale(self.get_scale())
                # # scale_bar.set_P1(self.get_P1().x()+extra_space, self.get_P1().y()+extra_space)
                # scale_bar.set_P1(self.get_P1())
                # alignRight(bottom_right, scale_bar)
                # alignBottom(bottom_right, scale_bar)
                # # scale_bar.set_P1(scale_bar.get_P1().x()-extra_space, scale_bar.get_P1().x()+extra_space)
                # scale_bar.drawAndFill(painter=painter)

                # add a bunch of inner objects that should be packed left if they exist, and some right and some top, etc
                # so that these objects are packed
                # maybe loop over them
                # could have as many text labels as desired
                # only one letter
                # as many insets as needed

                # # center 2
                # scale_bar = ScaleBar(411/2, '<font color="#FFFFFF">10µm</font>')
                # scale_bar.set_scale(self.get_scale())
                # # scale_bar.set_P1(self.get_P1().x()+extra_space, self.get_P1().y()+extra_space)
                # scale_bar.set_P1(self.get_P1())
                # alignCenterH(center, scale_bar)
                # alignCenterV(center, scale_bar)
                # # scale_bar.set_P1(scale_bar.get_P1().x()-extra_space, scale_bar.get_P1().x()+extra_space)
                # scale_bar.drawAndFill(painter=painter)

                # TODO create 5 reference points for each object and align to those
                # loop over all extra objects that need be added

    # def fill(self, painter, draw=True):
        # if self.fill_color is None:
        #     return
        # if draw:
        #     painter.save()
        # painter.setOpacity(self.opacity)
        # if draw:
        #     if self.img is not None:
        #         qsource = QRectF(0, 0, self.img.get_width(), self.img.get_height())
        #         painter.drawImage(self, self.qimage , qsource)
        #     else:
        #         painter.drawRect(self)
        #     painter.restore()
        # self.draw(painter=painter, draw=draw)

    # def drawAndFill(self, painter):
        # painter.save()
        # if self.img is not None:
        #     qsource = QRectF(0, 0, self.img.get_width(), self.img.get_height())
        #     painter.drawImage(self, self.qimage , qsource)
        # else:
        #     painter.drawRect(self)
        # painter.restore()
        # self.draw(painter=painter)

    # def __add__(self, other):
    def __or__(self, other):
        from epyseg.figure.row import Row  # KEEP Really required to avoid circular imports
        return Row(self, other)

    # create a Fig with divide
    # def __truediv__(self, other):
    def __truediv__(self, other):
        from epyseg.figure.column import Column  # KEEP Really required to avoid circular imports
        return Column(self, other)

    def __floordiv__(self, other):
        return self.__truediv__(other=other)

    # ai je vraiment besoin de ça ? en fait le ratio suffit et faut aussi que j'intègre le crop sinon va y avoir des erreurs
    # Force the montage width to equal 'width_in_px'
    def setToWidth(self, width_in_px):
        # pure_image_width = self.width()
        # ratio = width_in_px / pure_image_width
        # self.setWidth(width_in_px)
        # self.setHeight(self.height() * ratio)
        # # we recompute the scale for the scale bar # TODO BE CAREFUL IF EXTRAS ARE ADDED TO THE OBJECT AS THIS WOULD PERTURB THE COMPUTATIONS
        # self.update_scale()
        pure_image_width = self.width(scaled=False)# need original height and with in fact
        # if self.__crop_left is not None:
        #     pure_image_width -= self.__crop_left
        # if self.__crop_right is not None:
        #     pure_image_width -= self.__crop_right
        scale = width_in_px / pure_image_width
        self.scale = scale


    def setToHeight(self, height_in_px):
        # pure_image_height = self.height()
        # self.setHeight(height_in_px)
        # ratio = height_in_px / pure_image_height
        # self.setWidth(self.width() * ratio)
        # # we recompute the scale for the scale bar # TODO BE CAREFUL IF EXTRAS ARE ADDED TO THE OBJECT AS THIS WOULD PERTURB THE COMPUTATIONS
        # self.update_scale()
        pure_image_height = self.height(scaled=False)
        # if self.__crop_top is not None:
        #     pure_image_height-=self.__crop_top
        # if self.__crop_bottom is not None:
        #     pure_image_height-=self.__crop_bottom
        scale = height_in_px/pure_image_height
        self.scale = scale
        # need update bounds
        # scale is ok

    # def update_scale(self):
    #     # we recompute the scale for the scale bar # TODO BE CAREFUL IF EXTRAS ARE ADDED TO THE OBJECT AS THIS WOULD PERTURB THE COMPUTATIONS
    #     self.scale = self.get_scale()

    # def get_scale(self):
    #     # we recompute the scale for the scale bar # TODO BE CAREFUL IF EXTRAS ARE ADDED TO THE OBJECT AS THIS WOULD PERTURB THE COMPUTATIONS
    #     return self.width() / self.img.get_width()

    def crop(self, left=None, right=None, top=None, bottom=None, all=None):
        # print(self.boundingRect())
        if left is not None:
            self.__crop_left = left
            # self.setWidth(self.img.get_width() - self.__crop_left)
        if right is not None:
            self.__crop_right = right
            # self.setWidth(self.img.get_width() - self.__crop_right)
        if top is not None:
            self.__crop_top = top
            # self.setHeight(self.img.get_height() - self.__crop_top)
        if bottom is not None:
            self.__crop_bottom = bottom
            # self.setHeight(self.img.get_height() - self.__crop_bottom)
        if all is not None:
            self.__crop_left = all
            self.__crop_right = all
            self.__crop_top = all
            self.__crop_bottom = all
            # self.setWidth(self.img.get_width() - self.__crop_left)
            # self.setWidth(self.img.get_width() - self.__crop_right)
            # self.setHeight(self.img.get_height() - self.__crop_top)
            # self.setHeight(self.img.get_height() - self.__crop_bottom)

        # see how to crop actually because I need to create a qimage
        # self.qimage = self.img.crop()
        # print(self.boundingRect())
    # def set_to_scale(self, factor):
    #     self.scale = factor

    def set_to_translation(self, translation):
        self.translation = translation

    # WHY 0,0 FOR BOUNDS ??? --> sounds not smart to me
    def boundingRect(self, scaled=True):
 # en fait pas good besoin de prendre les crops et le scale en compte
        # is
        # rect_to_plot = self.adjusted(self.__crop_left, self.__crop_top, -self.__crop_right, -self.__crop_bottom)
        rect_to_plot = self.adjusted(0, 0, -self.__crop_right-self.__crop_left, -self.__crop_bottom-self.__crop_top)
        # rect_to_plot = self.adjusted(-self.__crop_left, -self.__crop_top, -self.__crop_right, -self.__crop_bottom)
        # rect_to_plot = self.adjusted(0,0,0,0)
        # print('begin rect_to_plot', rect_to_plot, self.scale)
        # if kwargs['draw']==True or kwargs['fill']==True:
        # if self.scale is None or self.scale==1:
        #     painter.drawRect(self)
        # else:
        # on clone le rect
        if self.scale is not None and self.scale != 1 and scaled:
            # TODO KEEP THE ORDER THIS MUST BE DONE THIS WAY OR IT WILL GENERATE PLENTY OF BUGS...
            new_width = rect_to_plot.width() * self.scale
            new_height = rect_to_plot.height() * self.scale
            # print(rect_to_plot.width(), rect_to_plot.height())  # here ok
            # setX changes width --> why is that

            # TODO BE EXTREMELY CAREFUL AS SETX AND SETY CAN CHANGE WIDTH AND HEIGHT --> ALWAYS TAKE SIZE BEFORE OTHERWISE THERE WILL BE A PB AND ALWAYS RESET THE SIZE WHEN SETX IS CALLED!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # Sets the left edge of the rectangle to the given x coordinate. May change the width, but will never change the right edge of the rectangle. --> NO CLUE WHY SHOULD CHANGE WIDTH THOUGH BUT BE CAREFUL!!!
            # rect_to_plot.setX(rect_to_plot.x() * self.scale)
            # rect_to_plot.setY(rect_to_plot.y() * self.scale)
            # maybe to avoid bugs I should use translate instead rather that set x but ok anyways
            # print(rect_to_plot.width(), rect_to_plot.height())# bug here --> too big

            # print(new_height, new_height, self.width(), self.scale, self.scale* self.width())
            rect_to_plot.setWidth(new_width)
            rect_to_plot.setHeight(new_height)
        return rect_to_plot

    # def set_P1(self, *args):
    #     if not args:
    #         logger.error("no coordinate set...")
    #         return
    #     if len(args) == 1:
    #         self.moveTo(args[0].x(), args[0].y())
    #     else:
    #         self.moveTo(QPointF(args[0], args[1]))

    def get_P1(self):
        return self.boundingRect().topLeft()

    def width(self, scaled=True):
        return self.boundingRect(scaled=scaled).width()

    def height(self, scaled=True):
        return self.boundingRect(scaled=scaled).height()

    SVG_INKSCAPE = 96
    SVG_ILLUSTRATOR = 72

    # NB THIS CODE IS BASED ON THE EZFIG SAVE CODE --> ANY CHANGE MADE HERE MAY ALSO BE MADE TO THE OTHER
    # best here would be for image to keep original size if nothing is specified
    # could also make it return a qimage for display in pyTA
    # qualities = [QPainter.NonCosmeticDefaultPen,  QPainter.SmoothPixmapTransform, QPainter.TextAntialiasing,QPainter.Antialiasing,QPainter.HighQualityAntialiasing]
    qualities = [  QPainter.SmoothPixmapTransform, QPainter.TextAntialiasing,QPainter.Antialiasing]

    def save(self, path, filetype=None, title=None, description=None, svg_dpi=SVG_INKSCAPE, quality=qualities[-1]):
        # if path is None or not isinstance(path, str):
        #     logger.error('please provide a valide path to save the image "' + str(path) + '"')
        #     return
        if path is None:
            filetype = '.tif'

        if filetype is None:
            if path.lower().endswith('.svg'):
                filetype = 'svg'
            else:
                filetype = os.path.splitext(path)[1]
        dpi = 72  # 300 # inkscape 96 ? check for illustrator --> check

        if filetype == 'svg':
            generator = QSvgGenerator()
            generator.setFileName(path)
            if svg_dpi == self.SVG_ILLUSTRATOR:
                generator.setSize(QSize(595, 842))
                generator.setViewBox(QRect(0, 0, 595, 842))
            else:
                generator.setSize(QSize(794, 1123))
                generator.setViewBox(QRect(0, 0, 794, 1123))

            if title is not None and isinstance(title, str):
                generator.setTitle(title)
            if description is not None and isinstance(description, str):
                generator.setDescription(description)
            generator.setResolution(
                svg_dpi)  # fixes issues in inkscape of pt size --> 72 pr illustrator and 96 pr inkscape but need change size

            painter = QPainter(generator)

            # print(generator.title(), generator.heightMM(), generator.height(), generator.widthMM(),
            #       generator.resolution(), generator.description(), generator.logicalDpiX())
        else:
            scaling_factor_dpi = 1
            # scaling_factor_dpi = self.scaling_factor_to_achieve_DPI(300)

            # in fact take actual page size ??? multiplied by factor
            # just take real image size instead


            # image = QtGui.QImage(QSize(self.cm_to_inch(21) * dpi * scaling_factor_dpi, self.cm_to_inch(29.7) * dpi * scaling_factor_dpi), QtGui.QImage.Format_RGBA8888) # minor change to support alpha # QtGui.QImage.Format_RGB32)

            # NB THE FOLLOWING LINES CREATE A WEIRD ERROR WITH WEIRD PIXELS DRAWN some sort of lines NO CLUE WHY

            img_bounds = self.boundingRect()
            image = QtGui.QImage(QSize(img_bounds.width() * scaling_factor_dpi, img_bounds.height()* scaling_factor_dpi),  QtGui.QImage.Format_RGBA8888)  # minor change to support alpha # QtGui.QImage.Format_RGB32)
            # print('size at dpi',QSize(img_bounds.width() * scaling_factor_dpi, img_bounds.height()* scaling_factor_dpi))
            # QSize(self.cm_to_inch(0.02646 * img_bounds.width())
            # self.cm_to_inch(0.02646 * img_bounds.height())
            # need convert pixels to inches
            # is there a rounding error

            # force white bg for non jpg
            try:
                # print(filetype.lower())
                # the tif and png file formats support alpha
                if not filetype.lower() == '.png' and not filetype.lower() == '.tif' and not filetype.lower() == '.tiff':
                    image.fill(QColor.fromRgbF(1,1,1))
                else:
                    # image.fill(QColor.fromRgbF(1, 1, 1, alpha=1))
                    # image.fill(QColor.fromRgbF(1, 1, 1, alpha=1))
                    # TODO KEEP in fact image need BE FILLED WITH TRANSPARENT OTHERWISE GETS WEIRD DRAWING ERRORS
                    # TODO KEEP SEE https://stackoverflow.com/questions/13464627/qt-empty-transparent-qimage-has-noise
                    # image.fill(qRgba(0, 0, 0, 0))
                    image.fill(QColor.fromRgbF(0,0,0,0))
            except:
                pass
            painter = QPainter(image)  # see what happens in case of rounding of pixels
            # painter.begin()
            painter.scale(scaling_factor_dpi, scaling_factor_dpi)
        painter.setRenderHint(quality)  # to improve rendering quality
        self.draw(painter)
        painter.end()
        if path is None:
            return image
        if filetype != 'svg':
            image.save(path)
            return image

    #based on https://stackoverflow.com/questions/19902183/qimage-to-numpy-array-using-pyside
    def convert_qimage_to_numpy(self, qimage):
        qimage = qimage.convertToFormat(QtGui.QImage.Format.Format_RGB32)
        width = qimage.width()
        height = qimage.height()
        image_pointer = qimage.bits() # creates a deep copy --> this is what I want
        image_pointer.setsize(qimage.byteCount())
        # arr = np.array(image_pointer,copy=True).reshape(height, width, 4)
        arr = np.array(image_pointer).reshape(height, width, 4)
        arr = arr[..., 0:3]
        arr = Img.RGB_to_BGR(arr)  # that seems to do the job
        return arr

    # simple equation for an image
    # or maybe ask for starting expression
    # autoincrement math to get the right value
    # def _get_width_equation(self):
    #     # incr = fig_tools.common_value
    #     # AR = self.width(False)/self.height(False)
    #     # variable =
    #     equa_w = Symbol('a'+str(id(self)))*self.width(False)+0
    #     # equa_h = variable*AR*self.width(False)+0
    #     # incr+=1
    #     # fig_tools.common_value = incr
    #     return equa_w #, equa_h

    # def _get_height_equation(self):
    #     # incr = fig_tools.common_value
    #     AR = self.width(False)/self.height(False)
    #     # equa_w = Symbol('a'+str(id(self)))*self.width(False)+0
    #     equa_h = Symbol('a'+str(id(self)))*AR*self.width(False)+0
    #     # incr+=1
    #     # fig_tools.common_value = incr
    #     return equa_h #, equa_h

    # use that so that I can expand
    # if unique --> do nothing if many see how to handle that???
    # see how
    def get_equation(self):
        from sympy import nsolve, exp, Symbol
        AR = self.width(False)/self.height(False)
        equa_w = Symbol('a'+str(id(self)))*self.width(False)+0
        equa_h = Symbol('a'+str(id(self)))*AR*self.width(False)+0
        return [equa_w] , [equa_h], None # equa in width, equa in height, equas to solve, not useful here though

    # allow to deepcopy and copy
    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

if __name__ == '__main__':
    # ça marche --> voici deux examples de shapes
    test = Image2D(x=12, y=0, width=100, height=100)  # could also be used to create empty image with

    print(test.img)
    print(test.boundingRect())
    print(test.get_P1().x())

    # bug qd on definit une image comme param
    # test = Image2D('./../data/counter/06.png')
    test = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/01.png')
    print(test.boundingRect())  # --> it is ok there so why not below # not callable --> why -->
    print(test.get_P1())  # ça marche donc où est le bug
    print(test.get_P1().y())  # ça marche donc où est le bug
    # print(test.getP1().width())
    # ça marche

    # try draw on the image the quivers
    # img0.setLettering('<font color="red">A</font>')
    # # letter
    # img0.annotation.append(Rect2D(88, 88, 200, 200, stroke=3, color=0xFF00FF))
    # img0.annotation.append(Ellipse2D(88, 88, 200, 200, stroke=3, color=0x00FF00))
    # img0.annotation.append(Circle2D(33, 33, 200, stroke=3, color=0x0000FF))
    test.annotation.append(Line2D(33, 33, 88, 88, stroke=3, color=0x0000FF))

    test.annotation.append(Line2D(128, 33, 88, 88, stroke=0.65, color=0xFFFF00))
    # img0.annotation.append(Freehand2D(10, 10, 20, 10, 20, 30, 288, 30, color=0xFFFF00, stroke=3))
    # # img0.annotation.append(PolyLine2D(10, 10, 20, 10, 20, 30, 288, 30, color=0xFFFF00, stroke=3))
    # img0.annotation.append(Point2D(128, 128, color=0xFFFF00, stroke=6))

    # add a save to the image --> so that it exports as a raster --> TODO


    # painter =
    # img = test.draw() # if no painter --> create one as not to lose any data and or allow to save as vectorial


    img = test.save('/E/Sample_images/sample_images_PA/trash_test_mem/mini_vide/analyzed/trash/test_line2D.tif')

    test.save('/E/Sample_images/sample_images_PA/trash_test_mem/mini_vide/analyzed/trash/test_line2D.svg')

    #trop facile --> just hack it so that it can return a single qimage # or return a numpy image that is then plotted -> should not be too hard !!! I think --> TODO
    img = test.convert_qimage_to_numpy(img) # --> ok but I just need to swap the channels then I'll be done --> try that maybe with plenty of input images just to see


    # shall I save
    # try with non RGB images just to see

    # img = Img.RGB_to_BGR(img)

    # almost there --> just need to check that the size of the image is ok and that everything is fine
    plt.imshow(img)
    plt.show()


    print('here')
    img2 = Image2D('/E/Sample_images/sample_images_PA/trash_test_mem/counter/02.png')
    print('here3')
    preview(img2)

    # test replace the plot of pyTA of the polarity by this one --> should be quite easy TODO I think --> TODO

    # ok --> it all seems to work --> see how I can handle that



    # --> all seems ok now
    # --> put this in the advanced sql plotter




# Here I will gather all the unit tests for GUIs
# make sure all runs with pyside, pyside2, pyqt5 and pyqt6
import unittest
import os
import sys
import warnings
import random
from epyseg.settings.global_settings import print_default_qtpy_UI_really_in_use, list_UIs, set_UI, \
    force_qtpy_to_use_user_specified

class TestGUI(unittest.TestCase):

    # DEV COMMENT TO KEEP: uncomment the lines below to check one specific UI
    # UI = 'pyqt5'
    UI = 'pyqt6'
    # UI = 'pyside2'
    # UI = 'pyside6'
    GUIs = ['TissueAnalyzer',
            'EPySeg',
            # 'DeepTools', # deprecated
            'EZFIG_GUI',
            'scrollable_EZFIG',
            'PyQT_markdown',
            'GT_editor',
            'ListGUI',
            'Createpaintwidget',
            'scrollable_paint',
            'dualList',
            'tascrollablepaint',
            # gut_SEG, # not public yet
            # Branchpaint, # not public yet
            # TEM_SEG, # not public yet
            # ImageSelectionPaint, # not public yet
            ] # GUIs to test


    def test_qtpy_force(self):
        import os
        force_qtpy_to_use_user_specified()
        os.environ['QT_API']=self.UI
        self.assertEqual(os.environ['QT_API'],self.UI)
        from qtpy.QtWidgets import QWidget
        self.assertEqual( os.environ['QT_API'],self.UI)

    def test_GUIs(self):
        force_qtpy_to_use_user_specified()
        os.environ['QT_API']=self.UI

        from epyseg.epygui import EPySeg
        from epyseg.figure.ezfgui import EZFIG_GUI
        from epyseg.figure.scrollableEZFIG import scrollable_EZFIG
        from epyseg.gui.pyqtmarkdown import PyQT_markdown
        from epyseg.ta.GUI.GT_editor_GUI import GT_editor
        from epyseg.ta.GUI.list_gui import ListGUI
        from epyseg.ta.GUI.paint2 import Createpaintwidget
        from epyseg.ta.GUI.pyta import \
            TissueAnalyzer  # --> this force changes to pyQT6 here but why ???? --> works well outside of tests but not from inside --> really no clue why ???
        from epyseg.ta.GUI.scrollablepaint import scrollable_paint
        from epyseg.ta.GUI.stackedduallist import dualList
        from epyseg.ta.GUI.tascrollablepaint import tascrollablepaint
        # from personal.amrutha_gut.GUT_seg_GUI import gut_SEG
        # from personal.amrutha_gut.branchpaint import Branchpaint
        # from personal.laurence_had_neurons.TEM_seg_GUI import TEM_SEG
        # from personal.pyTA.GUI.display.imageSelectionPaint import ImageSelectionPaint

        from qtpy.QtWidgets import QApplication
        app = QApplication(sys.argv)
        print_default_qtpy_UI_really_in_use()
        self.assertEqual(os.environ['QT_API'], self.UI)
        # for gui in self.GUIs:
        for iii, gui in enumerate(self.GUIs):
            print(iii, gui) #  even though it only prints ta it seems to run for all
            # w = gui()
            w = locals()[gui]()
            w.show()
            self.assertTrue(w.isVisible())
            # if random.uniform(0, 1)<0.5:
            #     self.assertTrue(False)
        #


if __name__ == '__main__':
    # DEV COMMENT KEEP nothing below works --> must uncomment at the beginning of the code

    # import os
    # os.environ['QT_API']='pyqt5'
    # TestGUI.UI ='pyqt5'
    # TestGUI.UI ='pyqt6'
    # TestGUI.UI ='pyside6'
    # TestGUI.UI ='pyside2'
    # TestGUI.UI = os.environ.get('TEST_UI', TestGUI.UI)
    # ne marche pas... --> need change it directly at the tip of the code
    unittest.main()



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from epyseg.draw.shapes.rect2d import Rect2D
from qtpy.QtCore import QPointF, QRectF
# logger
from epyseg.tools.logger import TA_logger

logger = TA_logger()

# en fait le pack vaudrait mieux le faire là ça serait bcp plus utile
# keep this in mind: https://stackoverflow.com/questions/34827132/changing-the-order-of-operation-for-add-mul-etc-methods-in-a-custom-c
class group(Rect2D):

    def __init__(self, *args, space=3):
        self.images = []
        for img in args:
            #     if isinstance(img, Column):
            #         self.images += img.images
            #     else:
            self.images.append(img)
        # init super empty
        super(Rect2D, self).__init__()
        # self.isSet = True # required to allow dragging
        self.space = space

        # self.sameHeight(space=space)
        # self.widthInPixel = width
        # self.setToWidth(width)
        self.updateBoudingRect()

    def updateBoudingRect(self):
        '''updates the image bounding rect depending on content'''
        x = None
        y = None
        x2 = None
        y2 = None
        for img in self:
            topLeft = img.get_P1()
            if x is None:
                x = topLeft.x()
            if y is None:
                y = topLeft.y()
            x = min(topLeft.x(), x)
            y = min(topLeft.y(), y)

            # print(img, img.boundingRect(), type(img))
            # print(img, img.boundingRect(), type(img), img.boundingRect().height())

            if x2 is None:
                x2 = topLeft.x() + img.boundingRect().width()
            if y2 is None:
                y2 = topLeft.y() + img.boundingRect().height()
            x2 = max(topLeft.x() + img.boundingRect().width(), x2)
            y2 = max(topLeft.y() + img.boundingRect().height(), y2)

        self.setX(x)
        self.setY(y)
        self.setWidth(x2 - x)
        self.setHeight(y2 - y)

    def setOrigin(self, *args):
        self.set_P1(*args)

    def set_P1(self, *args):
        curP1 = self.get_P1()
        Rect2D.set_P1(self, *args)
        newP1 = self.get_P1()
        for img in self:
            img.translate(newP1.x() - curP1.x(), newP1.y() - curP1.y())

        self.updateBoudingRect()

    def translate(self, *args):
        if len(args) == 1:
            point = args[0]
            QRectF.translate(self, point.x(), point.y())
            for img in self:
                img.translate(point.x(), point.y())
        else:
            QRectF.translate(self, args[0], args[1])
            for img in self:
                img.translate(QPointF(args[0], args[1]))

    def __iter__(self):
        ''' Returns the Iterator object '''
        self._index = -1
        return self

    def __next__(self):
        ''' Returns the next image or panel in the row '''
        self._index += 1
        if self._index < len(self.images):
            return self.images[self._index]
        # End of Iteration
        raise StopIteration

    # Forces the block to be of width (width_in_px)
    def setToWidth(self, width_in_px):

        # make all content same height then compute width and incompressible width of each object in fact

        if width_in_px is None:
            return
        nb_cols = len(self)
        pure_image_width = (self.boundingRect().width()) - (nb_cols - 1.) * self.space - self.getIncompressibleWidth()
        # print(width_in_px, self.getIncompressibleWidth(), (nb_cols - 1.) * self.space)
        width_in_px -= self.getIncompressibleWidth() + (nb_cols - 1.) * self.space
        ratio = width_in_px / pure_image_width
        for img in self:
            img.setToHeight(img.boundingRect().width() * ratio)


        self.packX(self.space)
        self.updateBoudingRect()


    # compute corresponding row height ???? so that it fits not so easy in fact

    # pb here l'incompressible space est le courant plus la somme des incompressible spaces du contenu
    # def setToWidth(self, width_in_px):
    #
    #     # make all content same height then compute width and incompressible width of each object in fact
    #
    #     if width_in_px is None:
    #         return
    #     nb_cols = len(self)
    #     # pure_image_height = self.boundingRect().height() - self.getIncompressibleHeight()
    #     # incompressible_width =  self.getIncompressibleWidth()
    #
    #     # for img in self:
    #     #     incompressible_width+=img.getIncompressibleWidth()
    #
    #     pure_image_width = (self.boundingRect().width()) -self.getIncompressibleWidth(), (nb_cols - 1.) * self.space
    #
    #     # print(width_in_px, self.getIncompressibleWidth(), (nb_cols - 1.) * self.space)
    #     width_in_px -= self.getIncompressibleWidth(), (nb_cols - 1.) * self.space
    #     # height_in_px -= self.getIncompressibleHeight()
    #     ratio = width_in_px / pure_image_width
    #     for img in self:
    #         print('target width before', width_in_px, img.boundingRect(), ratio)
    #         img.setToWidth(img.boundingRect().width() * ratio)
    #         print('target width after', width_in_px, img.boundingRect())
    #
    #
    #     self.packX(self.space)
    #     self.updateBoudingRect()

    # Forces the block to be of width (width_in_px)
    # tjrs faire ça row by row
    def setToHeight(self, height_in_px):
        # celui la marche mais celui du dessus ne marche pas --> pkoi ???

        if height_in_px is None:
            return
        nb_cols = len(self) #ce ne sera pas tjrs correct --> faudrait faire mieux que ça en fait
        # pure_image_width = (self.boundingRect().width()) -self.getIncompressibleWidth(), (nb_cols - 1.) * self.space
        # incompressible_width = self.getIncompressibleWidth()

        # for img in self:
        #     incompressible_width += img.getIncompressibleWidth()
        pure_image_width = (self.boundingRect().width()) - self.getIncompressibleWidth(), (nb_cols - 1.) * self.space

        pure_image_height = self.boundingRect().height()- self.getIncompressibleHeight()
        # print(height_in_px, self.getIncompressibleHeight(), (nb_cols - 1.) * self.space)
        height_in_px -= self.getIncompressibleHeight()
        ratio = height_in_px / pure_image_height
        for img in self:
            img.setToHeight(img.boundingRect().height() * ratio)

        print('values resize', pure_image_width, pure_image_height, height_in_px, ratio)

        self.packX(self.space)
        self.updateBoudingRect()

    def __len__(self):
        if self.images is None:
            return 0
        return len(self.images)

     # @return the block incompressible width
    def getIncompressibleWidth(self):
        return 0
        #     # in fact need get nb of cols a
        #     nb_cols = len(self)
        #     extra_space = (nb_cols - 1.) * self.space
        #     extra_space += self.getExtraIncompressibleWidth()
        #     return extra_space
        #
    # def getExtraIncompressibleWidth(self):
    #     extra_space = 0
    #     return extra_space
    #
    # # @return the block incompressible height
    def getIncompressibleHeight(self):
        return 0
        #     # in fact need get nb of cols a
        #     nb_rows= len(self)
        #     extra_space = (nb_rows - 1.) * self.space
        #     extra_space += self.getExtraIncompressibleHeight()
        #     return extra_space
    #
    # def getExtraIncompressibleHeight(self):
    #     extra_space = 0
    #     return extra_space

    def packX(self, space=3):
        last_x = 0
        last_y = 0

        for i in range(len(self.images)):
            img = self.images[i]
            if i != 0:
                last_x += space
            img.set_P1(last_x, img.get_P1().y())
            last_x = img.boundingRect().x() + img.boundingRect().width()

        self.updateBoudingRect()

    def packY(self, space=3):
        last_x = 0
        last_y = 0

        for i in range(len(self.images)):
            img = self.images[i]
            if i != 0:
                last_y += space
            img.set_P1(img.get_P1().x(), last_y)
            # get all the bounding boxes and pack them with desired space in between
            # get first point and last point in x
            x = img.boundingRect().x()
            y = img.boundingRect().y()
            last_x = img.boundingRect().x() + img.boundingRect().width()
            last_y = img.boundingRect().y() + img.boundingRect().height()
        self.updateBoudingRect()

    def alignTop(self, updateBounds=False):
        first_left = None
        for img in self:
            cur_pos = img.get_P1()
            if first_left is None:
                first_left = cur_pos
            img.set_P1(cur_pos.x(), first_left.y())
        if updateBounds:
            self.updateMasterRect()


import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy import QtWidgets
from qtpy.QtCore import QPoint, QPointF, Qt, QRectF
from qtpy.QtGui import QBrush, QPen, QColor, QTransform

from epyseg.tools.logger import TA_logger

logger = TA_logger()

class Ellipse2D(QtWidgets.QGraphicsEllipseItem):

    isSet = False

    def __init__(self, *args, color=0xFFFF00, fill_color=None, opacity=1., stroke=0.65, line_style=None,theta=0, **kwargs):
        super(Ellipse2D, self).__init__(*args)
        if not args:
            self.isSet = False
        else:
            self.isSet = True
        self.setRect(self.rect())
        self.color = color
        self.fill_color = fill_color
        self.stroke = stroke
        self.opacity = opacity
        self.scale = 1
        self.translation = QPointF()
        self.line_style = line_style
        # rotation
        self.theta = theta

    def set_rotation(self, theta):
        self.theta = theta

    def set_opacity(self, opacity):
        self.opacity = opacity

    def set_line_style(self,style):
        """
        allows lines to be dashed or dotted or have custom pattern
        :param style: a list of numbers or any of the following Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine but not Qt.CustomDashLine, Qt.CustomDashLine is assumed by default if a list is passed in. None is also a valid value that resets the line --> assume plain line
        :return:
        """
        self.line_style = style
        # if style is a list then assume custom pattern otherwise apply solidline

    def draw(self, painter, draw=True):
        if self.color is None and self.fill_color is None:
            return

        if draw:
            painter.save()
            painter.setOpacity(self.opacity)
        if self.color is not None:
            pen = QPen(QColor(self.color))
            if self.stroke is not None:
                pen.setWidthF(self.stroke)
            if self.line_style is not None:
                if self.line_style in [Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine]:
                    pen.setStyle(self.line_style)
                elif isinstance(self.line_style, list):
                    pen.setStyle(Qt.CustomDashLine)
                    pen.setDashPattern(self.line_style)
            painter.setPen(pen)
        else:
            painter.setPen(Qt.NoPen)
        if self.fill_color is not None:
            painter.setBrush(QBrush(QColor(self.fill_color)))
        if draw:
            rect_to_plot = self.rect().adjusted(0, 0, 0, 0)
            if self.scale is not None and self.scale != 1:
                # TODO KEEP THE ORDER THIS MUST BE DONE THIS WAY OR IT WILL GENERATE PLENTY OF BUGS...
                new_width = rect_to_plot.width() * self.scale
                new_height = rect_to_plot.height() * self.scale
                # TODO BE EXTREMELY CAREFUL AS SETX AND SETY CAN CHANGE WIDTH AND HEIGHT --> ALWAYS TAKE SIZE BEFORE OTHERWISE THERE WILL BE A PB AND ALWAYS RESET THE SIZE WHEN SETX IS CALLED!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                # Sets the left edge of the rectangle to the given x coordinate. May change the width, but will never change the right edge of the rectangle. --> NO CLUE WHY SHOULD CHANGE WIDTH THOUGH BUT BE CAREFUL!!!
                rect_to_plot.setX(rect_to_plot.x() * self.scale)
                rect_to_plot.setY(rect_to_plot.y() * self.scale)
                rect_to_plot.setWidth(new_width)
                rect_to_plot.setHeight(new_height)
            if self.translation is not None:
                rect_to_plot.translate(self.translation)

            # if self.color is not None:
            #     painter.drawRect(rect_to_plot)
            # else:
            #     painter.fillRect(rect_to_plot, QColor(self.fill_color))
            if self.theta is not None and self.theta != 0:
                painter.translate(rect_to_plot.center())
                painter.rotate(self.theta)
                painter.translate(-rect_to_plot.center())

            painter.drawEllipse(rect_to_plot)
            painter.restore()




    # def fill(self, painter, draw=True):
    #     if self.fill_color is None:
    #         return
    #     if draw:
    #         painter.save()
    #     painter.setBrush(QBrush(QColor(self.fill_color)))
    #     painter.setOpacity(self.opacity)
    #     if draw:
    #         painter.drawEllipse(self.rect())
    #         painter.restore()
    #
    #     # TODO pb will draw the shape twice.... ---> because painter drawpolygon is called twice
    #
    # def drawAndFill(self, painter):
    #     painter.save()
    #     self.draw(painter, draw=False)
    #     self.fill(painter, draw=False)
    #     painter.drawEllipse(self.rect())
    #     painter.restore()

    def translate(self, translation):
        self.moveBy(translation.x(), translation.y())
        rect = self.rect()
        rect.translate(translation.x(), translation.y())
        self.setRect(rect)

    def add(self, *args):
        p1 = args[0]
        p2 = args[1]
        rect = self.rect()
        rect.setWidth(abs(p1.x()-p2.x()))
        rect.setHeight(abs(p1.y()-p2.y()))
        x = p2.x()
        y = p2.y()
        if p1.x() < p2.x():
            x = p1.x()
        if p1.y() < p2.y():
            y = p1.y()
        rect.setX(x)
        rect.setY(y)

        self.setRect(rect)
        self.isSet = True

    def boundingRect(self):
        # should I return the scaled version or the orig --> think about it...
        rect_to_plot = self.rect().adjusted(0, 0, 0, 0)
        try:
            # print('tada')
            if self.theta is not None and self.theta!=0:
                # print('entering')
                center = rect_to_plot.center()
                # print('entering2')
                t = QTransform().translate(center.x(), center.y()).rotate(self.theta).translate(-center.x(),
                                                                                                -center.y())
                # print('entering3')
                # self.setTransform(t)
                # self.transform()


                # transformed = QRectF(self.rect())
                # print('entering5', transformed)
                # self.resetTransform()
                # print('entering5', rect_to_plot)
                # print('entering4')
                transformed = t.mapRect(rect_to_plot)

                # self.setTransform(t)
                # self.transform()
                #
                # print(self.shape().boundingRect())
                #
                # # print(self.rect(), transformed)
                #
                # transformed = QRectF(self.shape().boundingRect())
                # # self.resetTransform()

                # not perfect but ok for now though --> bounds are not sharp at the edges upon rotation
                # print('entering45', transformed)
                return transformed
        except:
            pass
        return rect_to_plot

    def get_P1(self):
        return self.boundingRect().topLeft()

    def set_P1(self, point):
        rect = self.rect()
        width = rect.width()
        height = rect.height()
        rect.setX(point.x())
        rect.setY(point.y())
        # required due to setX changing width and sety changing height
        rect.setWidth(width)
        rect.setHeight(height)
        self.setRect(rect)

    def set_to_scale(self, factor):
        self.scale = factor

    def set_to_translation(self, translation):
        self.translation = translation


if __name__ == '__main__':
    # ça marche --> voici deux examples de shapes
    test = Ellipse2D(0, 0, 100, 100)
    # print(test.x(), test.y(), test.width(), test.height())
    print(test.contains(QPointF(50, 50)))
    print(test.contains(QPointF(15, 15)))
    print(test.contains(QPointF(-1, -1)))
    print(test.contains(QPointF(0, 0)))
    print(test.contains(QPointF(100, 100)))
    print(test.contains(QPointF(100, 100.1)))
    print(test.x())
    print(test.y())
    print(test.translate(QPoint(10, 10)))
    print(test.x())
    print(test.y())

    # p1 = test.p1()
    # print(p1.x(), p1.y())
    # p2 = test.p2()
    # print(p2.x(), p2.y())
    # print(test.arrow)
    # print(test.length()) # sqrt 2 --> 141
    # # if it's an arrow I can add easily all the stuff I need
    #
    # test = Rect2D(0, 0, 1, 1)
    # p1 = test.p1()
    # print(p1.x(), p1.y())
    # p2 = test.p2()
    # print(p2.x(), p2.y())
    # print(test.arrow)
    # import math
    # print(test.length() == math.sqrt(2))  # sqrt 2
    #
    # test2 = Rect2D()
    # p1 = test2.p1()
    # print(p1.x(), p1.y())
    # p2 = test2.p2()
    # print(p2.x(), p2.y())
    # print(test2.arrow)



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy import QtCore
from qtpy.QtCore import QPointF, Qt
from qtpy.QtGui import QPainter, QBrush, QPen, QImage, QColor, QPolygonF
# from qtpy.Qt
# from qtpy.Qt import (QPaintEngine, QPaintDevice,  QTransform, QBrush)
from epyseg.draw.shapes.polygon2d import Polygon2D

from epyseg.tools.logger import TA_logger

logger = TA_logger()

class PolyLine2D(Polygon2D):

    def __init__(self, *args, color=0xFFFF00, opacity=1., stroke=0.65, line_style=None, theta=0, fill_color=None,  **kwargs):
        super(PolyLine2D, self).__init__()
        points = []
        if len(args) > 0:
            if isinstance(args[0],tuple):
                for i in range(0, len(args)):
                    self.append(QPointF(args[i][0], args[i][1]))
            else:
                for i in range(0, len(args), 2):
                    self.append(QPointF(args[i], args[i+1]))
        self.color = color
        self.fill_color = fill_color
        self.stroke = stroke
        self.opacity = opacity
        self.isSet = False
        self.line_style = line_style
        # rotation
        self.theta = theta

    def set_rotation(self, theta):
            self.theta = theta

    def set_opacity(self, opacity):
        self.opacity = opacity

    def set_line_style(self, style):
        """
        allows lines to be dashed or dotted or have custom pattern

        :param style: a list of numbers or any of the following Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine but not Qt.CustomDashLine, Qt.CustomDashLine is assumed by default if a list is passed in. None is also a valid value that resets the line --> assume plain line
        :return:
        """
        self.line_style = style
        # if style is a list then assume custom pattern otherwise apply solidline

    def draw(self, painter, draw=True):
        if self.color is None: # and self.fill_color is None
            return

        if draw:
            painter.save()
            pen = QPen(QColor(self.color))
            if self.stroke is not None:
                pen.setWidthF(self.stroke)
            if self.line_style is not None:
                if self.line_style in [Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine]:
                    pen.setStyle(self.line_style)
                elif isinstance(self.line_style, list):
                    pen.setStyle(Qt.CustomDashLine)
                    pen.setDashPattern(self.line_style)
            if self.color is not None:
                painter.setPen(pen)
            else:
                painter.setPen(Qt.NoPen)
            if self.fill_color is not None:
                # print('coloring')
                brush = QBrush(QColor(self.fill_color))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

            painter.setOpacity(self.opacity)
            polyline_to_draw = self.translated(0, 0)
            if self.scale is not None and self.scale != 1:
                polyline_to_draw = self.__scaled()
            # print('mid rect_to_plot', rect_to_plot)
            if self.translation is not None:
                # rect_to_plot.setX(rect_to_plot.x()+self.translation.x())
                # rect_to_plot.setY(rect_to_plot.y()+self.translation.y())
                polyline_to_draw.translate(self.translation.x(), self.translation.y())
            # painter.drawPolygon(polygon_to_draw)
            if self.theta is not None and self.theta != 0:
                painter.translate(polyline_to_draw.boundingRect().center())
                painter.rotate(self.theta)
                painter.translate(-polyline_to_draw.boundingRect().center())

            # painter.drawPolygon(polyline_to_draw)
            painter.drawPolyline(polyline_to_draw)
            painter.restore()
    #
    # def fill(self, painter, draw=True):
    #     if self.fill_color is None:
    #         return
    #     if draw:
    #         painter.save()
    #     painter.setBrush(QBrush(QColor(self.fill_color)))
    #     painter.setOpacity(self.opacity)
    #     if draw:
    #         painter.drawPolyline(self)
    #         painter.restore()
    #
    # # TODO pb will draw the shape twice.... ---> because painter drawpolygon is called twice
    # def drawAndFill(self, painter):
    #     painter.save()
    #     self.draw(painter, draw=False)
    #     self.fill(painter, draw=False)
    #     painter.drawPolyline(self)
    #     painter.restore()
    #
    # def setP1(self, point):
    #     self.append(point)

    def add(self, *args, force=True):
        if self.count() > 1:
            self.remove(self.count()-1)
        self.append(args[1])
        self.isSet = True

    def listVertices(self):
        return [point for point in self]

    def __scaled(self):
        vertices = self.listVertices()
        scaled_poly = QPolygonF()
        for vx in vertices:
            vx.setX(vx.x()*self.scale)
            vx.setY(vx.y()*self.scale)
            scaled_poly.append(vx)
        return scaled_poly

if __name__ == '__main__':
    test = PolyLine2D(0, 0, 10, 0, 10, 20, 0, 20, 0, 0)
    print(test.count()) # marche pas --> pas ajouté
    print(test)

    hexagon = PolyLine2D()
    print(hexagon)
    hexagon.append(QPointF(10, 20))
    hexagon.append(QPointF(10, 30))
    hexagon.append(QPointF(20, 30))

    print(hexagon)
    print(hexagon.isEmpty())
    print(hexagon.count())
    print(hexagon.stroke)
    print(hexagon.get_stroke_size())
    # ça marche maintenant
    # print(hexagon.)
    # trop cool l'acces aux points
    for point in hexagon:
        print(point)

    print(hexagon.get_points())
    print(hexagon.contains(QPointF(10, 20)))
    print(hexagon.contains(QPointF(10, 21)))

    print(hexagon.isClosed()) #
    hexagon.append(QPointF(10, 20)) # closing the hexagon --> the last and first point should be the same
    print(hexagon.isClosed())  #
    #
    # image = QImage('./../data/handCorrection.png')
    # painter = QPainter()
    # painter.begin(image)
    # # painter.setOpacity(0.3);
    # painter.drawImage(0, 0, image)
    # painter.setPen(QtCore.Qt.blue)
    # painter.drawPolygon(hexagon)
    # hexagon.opacity = 0.7
    # painter.translate(10, 20)
    # hexagon.draw(painter) # ça marche pourrait overloader ça avec du svg
    # painter.end()
    #
    # # painter.save()
    # # painter.setCompositionMode(QtGui.QPainter.CompositionMode_Clear)
    # # painter.eraseRect(r)
    # # painter.restore()
    #
    # image.save('./../trash/test_pyQT_draw.png', "PNG");
    #
    # #pas mal TODO faire une classe drawsmthg qui dessine n'importe quelle forme que l'on lui passe avec des parametres de couleur, transparence, ...

    # tt marche aps mal ça va très vite




import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
#import argparse # later maybe allow arguments
from epyseg.ta.GUI.pyta import TissueAnalyzer
from qtpy.QtWidgets import QApplication
import sys
# import multiprocessing
# import platform

if __name__ == '__main__':
    # mac OSX fix to test --> this does not work at all in my hands...
    # if platform.system() == 'Darwin':
    #     multiprocessing.set_start_method('spawn')
    # multiprocessing.set_start_method('fork')
    # multiprocessing.freeze_support()
    # multiprocessing.set_start_method('spawn')

    # run the deep learning GUI
    # code can be run from the command line using 'python -m epyseg'
    app = QApplication(sys.argv)
    w = TissueAnalyzer()
    w.show()
    sys.exit(app.exec_())




# below is the java code of my selection --> I should inspire myself from it
# see the best way todo that
# or maybe allow to handle only elements of a given scope --> in a way that simplifies a lot the things !!! --> think about it
# if scope is images --> only offer images options, if scope is panel --> offer panel options
# if scope is Figure --> only offer figure options


# package GUIs;
#
# import MyShapes.Figure;
# import MyShapes.Montage;
# import MyShapes.MyImage2D;
# import MyShapes.PARoi;
# import MyShapes.Row;
# import MyShapes.MyTextRow;
# import java.awt.geom.Point2D;
# import java.util.ArrayList;
#
# /**
#  * This class deals with selections
#  *
#  * @author benoit aigouy
#  */
# public class OptimalSelection {
#
#     public static Object getSelectionAt(ArrayList<Object> ROIs, boolean goDeeper, Point2D.Double mouseClick, Object objectsAlreadySelected) {
#         Object sel = null;
#         if (ROIs == null || ROIs.isEmpty()) {
#             return null;
#         }
#         int size = ROIs.size();
#         /* we loop in reverse order */
#         loop1:
#         for (int i = size - 1; i >= 0; i--) {
#             Object object = ROIs.get(i);
#             if (object == null) {
#                 continue;
#             }
#             if (objectsAlreadySelected instanceof MyImage2D) {
#                 /* we are at the lowest level --> we go up again */
#                 if ((object instanceof Figure) || (object instanceof Row) || (object instanceof Montage)) {
#                     if (object instanceof PARoi) {
#                         if (((PARoi) object).contains(mouseClick)) {
#                             return object;
#                         }
#                     }
#                 }
#             }
#             /* go deeper --> if selected object is a figure --> return the row under the mouse */
#             if (objectsAlreadySelected instanceof Figure) {
#                 Row montageContainingClick = ((Figure) objectsAlreadySelected).getRowAtPosition(mouseClick);
#                 if (montageContainingClick instanceof Row) {
#                     return montageContainingClick;
#                 }
#             }
#             /* go deeper --> if selected object is a row --> return the montage under the mouse */
#             if (objectsAlreadySelected instanceof Row) {
#                 Montage montageContainingClick = ((Row) objectsAlreadySelected).getMontageAtPosition(mouseClick);
#                 if (montageContainingClick instanceof Montage) {
#                     return montageContainingClick;
#                 }
#             }
#             /* go deeper --> if selected object is a Montage --> return the image2D under the mouse */
#             if (objectsAlreadySelected instanceof Montage) {
#                 MyImage2D montageContainingClick = ((Montage) objectsAlreadySelected).getMyImage2DAtPosition(mouseClick);
#                 if (montageContainingClick instanceof MyImage2D) {
#                     return montageContainingClick;
#                 }
#             }
#             if (object instanceof PARoi) {
#                 if (((PARoi) object).contains(mouseClick)) {
#                     return object;
#                 }
#             }
#         }
#         return sel;
#     }
#
#     /**
#      * adds to the current selection the newly selected object
#      *
#      * @param ROIs
#      * @param breakOnFirstEncounter
#      * @param onlyAllowSameInstanceAsSelection
#      * @param mouseClick
#      * @param objectsAlreadySelected
#      * @return
#      */
#     public static Object addToSelection(ArrayList<Object> ROIs, boolean breakOnFirstEncounter, boolean onlyAllowSameInstanceAsSelection, Point2D.Double mouseClick, Object objectsAlreadySelected) {
#         Object newSel = getSelection(ROIs, breakOnFirstEncounter, onlyAllowSameInstanceAsSelection, mouseClick, objectsAlreadySelected);
# //        System.out.println("potentially newly selected object " + newSel + " " + objectsAlreadySelected);
#         /* if there is no cur sel return new sel as cur sel */
#         if (objectsAlreadySelected == null) {
#             return newSel;
#         }
#         if (newSel == null) {
#             return objectsAlreadySelected;
#         }
#         /* if sel is already a list of selected objects add the new sel to the list and return it */
#         if (objectsAlreadySelected instanceof ArrayList) {
#             ArrayList<Object> sel = ((ArrayList<Object>) objectsAlreadySelected);
#             if (!sel.contains(newSel)) {
#                 sel.add(newSel);
#             }
#             return sel;
#         } else {
#             /* there is already a selected object but we create a list and add all objects to it */
#             ArrayList<Object> sel = new ArrayList<Object>();
#             sel.add(objectsAlreadySelected);
#             if (newSel != objectsAlreadySelected) {
#                 sel.add(newSel);
#             }
#             return sel;
#         }
#     }
#
#     public static Object getSelection(ArrayList<Object> ROIs, boolean breakOnFirstEncounter, boolean onlyAllowSameInstanceAsSelection, Point2D.Double mouseClick, Object objectsAlreadySelected) {
#         if (ROIs == null || ROIs.isEmpty()) {
#             return null;
#         }
#         int size = ROIs.size();
#         ArrayList<Object> foundObjects = new ArrayList<Object>();
#         /* we loop in reverse order */
#         loop1:
#         for (int i = size - 1; i >= 0; i--) {
#             Object object = ROIs.get(i);
#             /* we prevent the object from being selected twice */
#             if (object == objectsAlreadySelected) {
#                 continue;
#             }
#             if (objectsAlreadySelected != null && (objectsAlreadySelected instanceof ArrayList)) {
#                 /* prevent from selecting the same object again */
#                 ArrayList<Object> curObjects = ((ArrayList<Object>) objectsAlreadySelected);
#                 for (Object foundObject : curObjects) {
#                     if (object == foundObject) {
#                         continue loop1;
#                     }
#                 }
#             }
#             if (object instanceof PARoi) {
#                 if (((PARoi) object).contains(mouseClick)) {
#                     if (objectsAlreadySelected == null) {
#                         return object;
#                     }
#                     if (onlyAllowSameInstanceAsSelection) {
#                         /* if they are of the same instance take it directly otherwise look inside */
#  /* here we just get objects of the same type */
#                         if ((object instanceof Figure) && ((objectsAlreadySelected instanceof Figure) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof Figure))) {
#                             if (breakOnFirstEncounter) {
#                                 return object;
#                             } else {
#                                 foundObjects.add(object);
#                             }
#                         }
#                         /* prevent MyTextRow to be selected with Rows even if they are Rows */
#                         if (((object instanceof MyTextRow)) && (((objectsAlreadySelected instanceof MyTextRow)) || ((objectsAlreadySelected instanceof ArrayList) && (((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof MyTextRow)))) {
#                             if (breakOnFirstEncounter) {
#                                 return object;
#                             } else {
#                                 foundObjects.add(object);
#                             }
#                         } else if (((object instanceof Row) && !(object instanceof MyTextRow)) && (((objectsAlreadySelected instanceof Row) && !(objectsAlreadySelected instanceof MyTextRow)) || ((objectsAlreadySelected instanceof ArrayList) && (((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof Row) && !(((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof MyTextRow)))) {
#                             if (breakOnFirstEncounter) {
#                                 return object;
#                             } else {
#                                 foundObjects.add(object);
#                             }
#                         }
#                         if ((object instanceof Montage) && ((objectsAlreadySelected instanceof Montage) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof Montage))) {
#                             if (breakOnFirstEncounter) {
#                                 return object;
#                             } else {
#                                 foundObjects.add(object);
#                             }
#                         }
#                         if ((object instanceof MyImage2D) && ((objectsAlreadySelected instanceof MyImage2D) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof MyImage2D))) {
#                             if (breakOnFirstEncounter) {
#                                 return object;
#                             } else {
#                                 foundObjects.add(object);
#                             }
#                         }
#                         if ((object instanceof Figure) && ((objectsAlreadySelected instanceof MyTextRow) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof MyTextRow))) {
#                             Object innerSelectedObject = ((Figure) object).getRowAtPosition(mouseClick);
#                             if (innerSelectedObject instanceof MyTextRow) {
#                                 if (breakOnFirstEncounter) {
#                                     return innerSelectedObject;
#                                 } else {
#                                     foundObjects.add(innerSelectedObject);
#                                 }
#                             }
#                         } else if ((object instanceof Figure) && (((objectsAlreadySelected instanceof Row) && !(objectsAlreadySelected instanceof MyTextRow)) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof Row && !(((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof MyTextRow)))) {
#                             Object innerSelectedObject = ((Figure) object).getRowAtPosition(mouseClick);
#                             if ((innerSelectedObject instanceof Row) && !(innerSelectedObject instanceof MyTextRow) ) {
#                                 if (breakOnFirstEncounter) {
#                                     return innerSelectedObject;
#                                 } else {
#                                     foundObjects.add(innerSelectedObject);
#                                 }
#                             }
#                         }
#                         /* if instance of objects differ then take the compatible one when grouping */
#                         if ((object instanceof Figure) && ((objectsAlreadySelected instanceof Montage) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof Montage))) {
#                             Object innerSelectedObject = ((Figure) object).getRowAtPosition(mouseClick);
#                             if (innerSelectedObject instanceof Row) {
#                                 innerSelectedObject = ((Row) innerSelectedObject).getMontageAtPosition(mouseClick);
#                             }
#                             if (innerSelectedObject instanceof Montage) {
#                                 if (breakOnFirstEncounter) {
#                                     return innerSelectedObject;
#                                 } else {
#                                     foundObjects.add(innerSelectedObject);
#                                 }
#                             }
#                         }
#                         /* if instance of objects differ then take the compatible one when grouping */
#                         if ((object instanceof Figure) && ((objectsAlreadySelected instanceof MyImage2D) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof MyImage2D))) {
#                             Object innerSelectedObject = ((Figure) object).getRowAtPosition(mouseClick);
#                             if (innerSelectedObject instanceof Row) {
#                                 innerSelectedObject = ((Row) innerSelectedObject).getMontageAtPosition(mouseClick);
#                             }
#                             if (innerSelectedObject instanceof Montage) {
#                                 innerSelectedObject = ((Montage) innerSelectedObject).getMyImage2DAtPosition(mouseClick);
#                             }
#                             if (innerSelectedObject instanceof MyImage2D) {
#                                 if (breakOnFirstEncounter) {
#                                     return innerSelectedObject;
#                                 } else {
#                                     foundObjects.add(innerSelectedObject);
#                                 }
#                             }
#                         }
#                         /* if instance of objects differ then take the compatible one when grouping */
#                         if ((object instanceof Row) && ((objectsAlreadySelected instanceof Montage) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof Montage))) {
#                             //get the montage at position
#                             Object innerSelectedObject = ((Row) object).getMontageAtPosition(mouseClick);
#                             if (innerSelectedObject instanceof Montage) {
#                                 if (breakOnFirstEncounter) {
#                                     return innerSelectedObject;
#                                 } else {
#                                     foundObjects.add(innerSelectedObject);
#                                 }
#                             }
#                         }
#                         /* if instance of objects differ then take the compatible one when grouping */
#                         if ((object instanceof Row) && ((objectsAlreadySelected instanceof MyImage2D) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof MyImage2D))) {
#                             Object innerSelectedObject = ((Row) object).getMontageAtPosition(mouseClick);
#                             if (innerSelectedObject instanceof Montage) {
#                                 innerSelectedObject = ((Montage) innerSelectedObject).getMyImage2DAtPosition(mouseClick);
#                             }
#                             if (innerSelectedObject instanceof MyImage2D) {
#                                 if (breakOnFirstEncounter) {
#                                     return innerSelectedObject;
#                                 } else {
#                                     foundObjects.add(innerSelectedObject);
#                                 }
#                             }
#                         }
#                         /* if instance of objects differ then take the compatible one when grouping */
#                         if ((object instanceof Montage) && ((objectsAlreadySelected instanceof MyImage2D) || ((objectsAlreadySelected instanceof ArrayList) && ((ArrayList<Object>) objectsAlreadySelected).get(0) instanceof MyImage2D))) {
#                             //get the montage at position
#                             Object innerSelectedObject = ((Montage) object).getMyImage2DAtPosition(mouseClick);
#                             if (innerSelectedObject instanceof MyImage2D) {
#                                 if (breakOnFirstEncounter) {
#                                     return innerSelectedObject;
#                                 } else {
#                                     foundObjects.add(innerSelectedObject);
#                                 }
#                             }
#                         }
#                     } else {
#                         return object;
#                     }
#                 }
#             }
#         }
#         if (!foundObjects.isEmpty()) {
#             return foundObjects;
#         }
#         return null;
#     }
#
#     public static void main(String[] args) {
#         System.exit(0);
#     }
#
# }
